#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-06
# @Filename: skies.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import multiprocessing
import os
import warnings
from functools import partial

from typing import Literal, Sequence, overload

import enlighten
import healpy
import numpy
import numpy.typing as npt
import pandas
import peewee
import polars
from astropy import units as uu
from astropy.coordinates import (
    SkyCoord,
    match_coordinates_sky,
    search_around_sky,
)
from astropy.time import Time
from matplotlib import pyplot as plt
from matplotlib.patches import Ellipse
from mocpy import MOC
from pydantic import BaseModel, ConfigDict, model_validator
from scipy.stats import circmean
from typing_extensions import Self

from sdssdb.connection import PeeweeDatabaseConnection

from target_selection import log
from target_selection.exceptions import TargetSelectionError, TargetSelectionUserWarning


warnings.filterwarnings("ignore", ".*invalid value encountered in power.*")


_known_flux_zpts = {
    "nMgy": 22.5,
    "Jy": 8.9,
    "mJy": 16.4,
    "uJy": 23.9,
}


class CatalogueParams(BaseModel):
    """A model to hold the parameters for a catalogue."""

    model_config = ConfigDict(extra="forbid")

    ra_column: str
    dec_column: str
    pmra_column: str | None = None
    pmdec_column: str | None = None
    ref_epoch: float | str | None = None
    mag_column: str | None = None
    radius_column: str | None = None
    is_flux: bool = False
    flux_unit: str = "nMgy"
    mag_threshold: float = 14.0
    min_separation: float = 5.0
    scale_a: float = 0.2
    scale_b: float = 1.5

    @model_validator(mode="after")
    def check_pm_fields(self) -> Self:
        """Check that all the pm fields are set or null."""

        if self.pmra_column is not None or self.pmdec_column is not None:
            assert (
                self.pmra_column is not None and self.pmdec_column is not None
            ), "both pmra_column and pmdec_column must be set."

            assert self.ref_epoch is not None, "ref_epoch must be set if pmra_column is set."

        return self


DEFAULT_CATALOGUE_PARAMS = {
    "twomass_psc": {
        "ra_column": "ra",
        "dec_column": "decl",
        "mag_column": "h_m",
        "mag_threshold": 14.0,
        "min_separation": 5.0,
        "scale_a": 0.2,
        "scale_b": 1.5,
    },
    "gaia_dr2_source": {
        "ra_column": "ra",
        "dec_column": "dec",
        "pmra_column": "pmra",
        "pmdec_column": "pmdec",
        "ref_epoch": 2015.5,
        "mag_column": "phot_g_mean_mag",
        "mag_threshold": 14.0,
        "min_separation": 5.0,
        "scale_a": 0.2,
        "scale_b": 1.5,
    },
    "gaia_dr3_source": {
        "ra_column": "ra",
        "dec_column": "dec",
        "pmra_column": "pmra",
        "pmdec_column": "pmdec",
        "ref_epoch": 2016,
        "mag_column": "phot_g_mean_mag",
        "mag_threshold": 14.0,
        "min_separation": 5.0,
        "scale_a": 0.2,
        "scale_b": 1.5,
    },
    "panstarrs1": {
        "ra_column": "ra",
        "dec_column": "dec",
        "mag_column": "r_stk_psf_flux",
        "is_flux": True,
        "flux_unit": "Jy",
        "mag_threshold": 14.0,
        "min_separation": 5.0,
        "scale_a": 0.2,
        "scale_b": 1.5,
    },
    "legacy_survey_dr8": {
        "ra_column": "ra",
        "dec_column": "dec",
        "pmra_column": "pmra",
        "pmdec_column": "pmdec",
        "ref_epoch": "ref_epoch",
        "mag_column": "flux_r",
        "is_flux": True,
        "flux_unit": "nMgy",
        "mag_threshold": 14.0,
        "min_separation": 5.0,
        "scale_a": 0.2,
        "scale_b": 1.5,
    },
    "tycho2": {
        "ra_column": "ramdeg",
        "dec_column": "demdeg",
        "pmra_column": "pmra",
        "pmdec_column": "pmde",
        "ref_epoch": 2000.0,
        "mag_column": "vtmag",
        "mag_threshold": 14.0,
        "min_separation": 5.0,
        "scale_a": 0.2,
        "scale_b": 1.5,
    },
    "twomass_xsc": {
        "ra_column": "ra",
        "dec_column": "decl",
        "radius_column": "r_ext",
    },
}


def nested_regrade(pixels: int | numpy.ndarray, nside_in: int, nside_out: int):
    r"""Returns the parent/children pixels from a given HealPix nested pixel.

    The HealPix nested mode follows a quadrilateral tree pixel scheme (see
    Figure 1 in Górski et al. 2005) When the resolution (nside) increases,
    each pixel is divided in four pixels. The numbering of such pixels follows
    a simple binary representation achieved by appending binary digits to the
    new pixels. So, for example, pixel 22 (``b10110``) becomes four new pixels
    with binary numbers ``b1011000, b1011001, b1011010, and b1011011``,
    corresponding to decimal ``91, 92, 93, and 94``.

    This function returns the child pixels from a given pixel when going from
    resolution ``nside_in`` to ``nside_out``, if ``nside_out > nside_in``, or
    the parent pixel if ``nside_out < nside_in``.

    Note that this function works only for pixels using the nested indexing and
    should not be used with the ring indexing.

    Parameters
    ----------
    pixels : int or ~numpy.ndarray
        The pixels for which we want to get the parents/children. Can be a
        single integer or an array of indices.
    nside_in : int
        The ``nside`` of the input pixels. Must be one of :math:`2^k` where
        :math:`k\in[0,1,2,\dots]`.
    nside_out : int
        The destination ``nside``.

    Returns
    -------
    output : `int` or `~numpy.ndarray`
        If ``nside_out < nside_in`` and ``pixels`` is a single value, this
        will be an integer with the parent of the input pixel. If ``pixels``
        is an array the output will be an array of the same size in which each
        element is the parent of the corresponding pixel in the input array.
        If ``nside_out > nside_in`` and ``pixels`` is an integer, the output
        will be an array with the child pixels. The size of the array will be
        :math:`2^{2l}` where ``l=k_out-k_in`` (:math:`n_{side}=2^k`).
        If the input in an array of pixels, the output will be a 2D array in
        which each row contains the child pixels of the corresponding pixel
        in the input array.

    """

    assert nside_in != nside_out, "nside_in cannot be equal to nside_out."

    k_in = numpy.log2(nside_in)
    k_out = numpy.log2(nside_out)

    assert k_in.is_integer() and k_out.is_integer(), "nside_in or nside_out are not power of 2."

    pixels = numpy.atleast_1d(pixels).astype(int)
    assert pixels.ndim == 1, "dimension of input pixels is invalid."

    # Npix = 12 * nside**2
    is_all = numpy.all(pixels <= 12 * nside_in**2)
    assert is_all, "some pixel indices are greater than the maximum allowed for nside_in."

    if k_in > k_out:
        degraded = numpy.right_shift(pixels, 2 * int(k_in - k_out))

        return degraded[0] if len(degraded) == 1 else degraded

    else:
        prograded = numpy.zeros((len(pixels), int(2 ** (2 * (k_out - k_in)))), dtype=int)
        prograded += pixels[numpy.newaxis].T
        for ii in range(int(k_out - k_in))[::-1]:
            prograded = numpy.left_shift(prograded, 2)

            repeats = numpy.repeat([0, 1, 2, 3], (4**ii) or 1)
            tiles = numpy.tile(repeats, prograded.shape[1] // len(repeats))

            prograded += tiles

        return prograded


def downsample(
    df: pandas.DataFrame,
    nsample: int = 2048,
    tile_column: str = "tile_32",
    tile_nside: int = 32,
    candidate_nside: int = 32768,
    downsample_nside: int = 256,
    downsample_data: pandas.DataFrame | None = None,
    seed: int | None = None,
):
    """Downsamples valid sky positions for a tile."""

    df["selected"] = False

    if "down_pix" not in df:
        df.loc[:, "down_pix"] = nested_regrade(df.index.values, candidate_nside, downsample_nside)

    if downsample_data is not None:
        cond = True
        if tile_column in df:
            tile = df.iloc[0][tile_column]
            cond = cond & (downsample_data[tile_column] == tile)
        if "valid" in df:
            cond = cond & downsample_data.valid
        down_tile = downsample_data.loc[cond]
        selected = df.reindex(down_tile.index).dropna()
        selected = selected.loc[selected.valid]
        df.loc[selected.index, "selected"] = True
        if len(selected) == nsample:
            return df

    # Step one: get as many valid sky positions as possible in random order.
    n_selected = df.selected.sum()
    n_valid_left = (df.valid & ~df.selected).sum()

    # Check if we are done.
    if n_selected >= nsample:
        return df
    elif (~df.selected).sum() == 0:
        return df

    # Calculate how many skies per downsample pixel to get.
    k_tile = int(numpy.log2(tile_nside))
    k_downsample = int(numpy.log2(downsample_nside))
    n_downsample_pix = 4 ** (k_downsample - k_tile)
    n_skies_per_pix = nsample // n_downsample_pix
    # reset index line here is new and it is meant to avoid problem in sample
    # with invalid ids while using weights
    # line with set_index here is not part of the original code and it is
    # meant to avoid problem in sample with invalid ids using weights
    if n_valid_left > 0:
        not_selected = df[(~df.selected) & df.valid]
        not_selected.reset_index(inplace=True)
        weights = not_selected.sep_neighbour / not_selected.sep_neighbour.sum()
        valid_selected = not_selected.groupby("down_pix").sample(
            n=n_skies_per_pix,
            replace=True,
            random_state=seed,
            weights=weights,
        )
        valid_selected.drop_duplicates(inplace=True)
        valid_selected.set_index("pix_32768", inplace=True)
        df.loc[valid_selected.index, "selected"] = True

    if df.selected.sum() >= nsample or (~df.selected).sum() == 0:
        return df

    # Step two: complete downsample pixels without enough skies with the
    # invalid skies that have the largest separation.

    assigned = df.groupby("down_pix").apply(lambda x: len(x[x.selected]))
    n_pix_missing = assigned[assigned < n_skies_per_pix]

    invalid_sorted = (
        df.loc[df.down_pix.isin(n_pix_missing.index) & ~df.selected]
        .groupby("down_pix")
        .apply(lambda x: x.sort_values(["valid", "sep_neighbour"], ascending=False))
    )

    if invalid_sorted.size == 0:
        return df

    if "down_pix" in invalid_sorted.index.names:
        invalid_sorted.reset_index("down_pix", drop=True, inplace=True)

    invalid_selected = invalid_sorted.groupby("down_pix").apply(
        lambda x: x.head(n_skies_per_pix - assigned.loc[x.iloc[0].down_pix])
    )

    if invalid_selected.size == 0:
        return df

    if "down_pix" in invalid_selected.index.names:
        invalid_selected.reset_index("down_pix", drop=True, inplace=True)

    df.loc[invalid_selected.index, "selected"] = True

    return df


def _process_tile(
    tile: int,
    database_params: dict,
    candidate_nside: int,
    tile_nside: int,
    query: str,
    min_separation: float,
    is_flux: bool,
    flux_unit: str,
    mag_threshold: float | None,
    scale_a: float,
    scale_b: float,
    nsample: int,
    downsample_data: pandas.DataFrame | None,
    downsample_nside: int,
    calculate_min_separation: bool,
    seed: int | None,
):
    """Processes a tile from catalogue data."""

    db_params = database_params.copy()
    database = db_params.pop("database")

    db = peewee.PostgresqlDatabase(database, **db_params)
    db.connect()

    targets = pandas.read_sql(query.format(tile=tile), db)  # type: ignore

    db.close()

    has_mag = "mag" in targets
    has_radius = "radius" in targets

    if "mag" in targets and not mag_threshold:
        raise TargetSelectionError("mag_threshold required if mag_column is set.")

    if len(targets) == 0:
        return False

    # For each target we calculate the corresponding pixel in candidate_nside
    # resolution.
    cpcol = f"pix_{candidate_nside}"

    targets[cpcol] = healpy.ang2pix(
        candidate_nside, targets.ra.array, targets.dec.array, nest=True, lonlat=True
    )

    # Get all the pixels in candidate_nside resolution that correspond to this
    # tile. Create a mask with all of them that are not occupied by a target.
    all_pix = nested_regrade(tile, tile_nside, candidate_nside)[0]
    mask_not_target = ~numpy.isin(all_pix, targets[cpcol])

    # Create a DF with the candidate pixels and calculate their RAs and Decs.
    candidates = pandas.DataFrame(all_pix[mask_not_target], columns=[cpcol])
    candidates["ra"], candidates["dec"] = healpy.pix2ang(
        candidate_nside,
        candidates[cpcol],
        nest=True,
        lonlat=True,
    )

    candidates["valid"] = True

    # If we are using the magnitudes to correct the minimum separation, we
    # first convert the column to magnitudes if it's flux. Then for each pixel
    # we mask out all the candidate pixels that lie within the corrected
    # separation.

    if is_flux:
        mask_flux = targets.mag > 0

        try:
            zpt = _known_flux_zpts[flux_unit]
        except BaseException:
            raise Exception(f"Unknown flux_unit: {flux_unit}")

        targets.loc[mask_flux, "mag"] = zpt - 2.5 * numpy.log10(targets[mask_flux].mag)
        targets.loc[targets.mag <= 0, "mag"] = numpy.nan

    vectors = healpy.pixelfunc.ang2vec(targets.ra, targets.dec, lonlat=True)

    all_masked = []
    row = 0
    for _, target in targets.iterrows():
        if has_radius:
            sep_corr = max(min_separation, target.radius)
        elif has_mag and mag_threshold is not None:
            min_sep_corr = numpy.power(mag_threshold - target.mag, scale_b) / scale_a
            sep_corr = min_separation + min_sep_corr
        else:
            sep_corr = min_separation

        masked = healpy.query_disc(
            candidate_nside,
            vec=vectors[row],
            radius=numpy.deg2rad(sep_corr / 3600.0),
            inclusive=True,
            fact=4,
            nest=True,
        )

        all_masked += masked.tolist()

        row += 1

    candidates.loc[candidates[cpcol].isin(all_masked), "valid"] = False

    # Determine the closest neighbout for each candidate pixel.
    if calculate_min_separation:
        c1 = SkyCoord(ra=candidates.ra.to_numpy(), dec=candidates.dec.to_numpy(), unit="deg")
        c2 = SkyCoord(ra=targets.ra.to_numpy(), dec=targets.dec.to_numpy(), unit="deg")

        target_idx, sep, _ = match_coordinates_sky(c1, c2, nthneighbor=1)

        matched_target = targets.iloc[target_idx, :]
        sep_arcsec = sep.value * 3600.0

        # Add column with the separation.
        candidates["sep_neighbour"] = sep_arcsec

        if has_mag:
            candidates["mag_neighbour"] = matched_target.loc[:, "mag"].to_numpy()

    candidates.loc[:, f"tile_{tile_nside}"] = tile

    # If we are not downsampling, set the index and return.
    if downsample is False or downsample is None:
        if len(candidates) > 0:
            candidates.set_index(cpcol, inplace=True)
        else:
            return False
        return candidates

    # If downsampling, regrade each candidate pixel value to the nside we'll
    # use for downsampling.
    candidates.loc[:, "down_pix"] = nested_regrade(
        candidates[cpcol],
        candidate_nside,
        downsample_nside,
    )

    if len(candidates) > 0:
        candidates.set_index(cpcol, inplace=True)
    else:
        return False

    if downsample_data is None and nsample is None:
        return candidates
    else:
        if downsample_data is not None:
            assert nsample is not None, "downsample_data requires nsample to be defined."
        downsampled = downsample(
            candidates,
            tile_column=f"tile_{tile_nside}",
            tile_nside=tile_nside,
            nsample=nsample,
            candidate_nside=candidate_nside,
            downsample_nside=downsample_nside,
            downsample_data=downsample_data,
            seed=seed,
        )
        return downsampled[downsampled.selected]


def get_sky_table(
    database,
    table,
    output,
    tiles=None,
    tile_nside=32,
    candidate_nside=32768,
    min_separation=10.0,
    ra_column="ra",
    dec_column="dec",
    mag_column=None,
    is_flux=False,
    radius_column=None,
    flux_unit="nMgy",
    scale_a=0.2,
    scale_b=1.0,
    mag_threshold=None,
    calculate_min_separation=True,
    nsample=2048,
    downsample_data=None,
    downsample_nside=256,
    n_cpus=1,
    seed=None,
):
    r"""Identifies skies from a database table.

    Skies are selected using the following procedure:

    - The sky is divided in HEALPix "tiles" of nside ``tile_nside``. For each
      tile the catalogue is queried to retrieve all the targets that lie in
      the tile footprint. This assumes that the ``healpix_ang2ipix_nest``
      Postgresql function from `pg_healpix
      <https://github.com/segasai/pg_healpix>`__ is available. It's recommended
      that an index is created for the tiling norder to speed the query.

    - Each tile is subsequently divided in pixels of nside ``candidate_nside``.
      Pixels that contain a catalogue target are removed as possible sky
      candidates. For each target with magnitude less than ``mag_threshold``
      all pixels within a ``min_separation`` are rejected. The remaining pixels
      are considered valid skies. Alternatively, if ``mag_column`` and
      ``mag_threshold`` are defined, the minimum separation to valid pixels is
      corrected using the expression
      :math:`s^* = s + \dfrac{(m_{thr}-m)^{\beta}}{a}` where
      :math:`s` is the minimum separation, :math:`m_{thr}` is the magnitude
      threshold, :math:`a=0.2` and :math:`\beta=1.0` are factors that
      control the relationship between the star's magnitude and the exclusion
      radius.

    - If ``nsample``, only that number of skies are returned for each tile.
      The tile is divided in pixels of nside ``downsample_nside`` (which must
      be smaller than ``candidate_nside`` but larger than ``tile_nside``) and
      for each downsample pixel ``int(downsample / downsample_npix) + 1`` skies
      are selected. If not enough valid positions can be selected in this way,
      each downsample pixel quota is completed with the invalid positions that
      have a larger separation to their nearest neighbour.

    - The process can be parallelised for each tile and the results are
      compiled in a single Pandas data frame that is saved to an HDF5 file.

    All pixel values use the nested ordering.

    Parameters
    ----------
    database : ~sdssdb.connection.PeeweeDatabaseConnection
        A valid database connection.
    table : str
        Name of the table to query, optionally schema-qualified.
    output : str
        Path to the HDF5 file to write.
    tiles : list
        A list of HEALPix pixels of nside ``tile_nside`` for which the sky
        selection will be done. If `None`, runs for all the pixels of nside
        ``tile_nside``.
    tile_nside : int
        The HEALPix nside to use to tile the all-sky catalogue.
    candidate_nside : int
        The HEALPix nside used to identify candidate pixels in each tile.
        Candidates are then checked to confirm that their closest neighbour
        is at least ``min_separation`` arcsec away.
    min_separation : int
        The minimum separation, in arcsec, between skies and their closest
        neighbour in the catalogue.
    ra_column : str
        The name of the column in ``table`` that contains the Right Ascension
        coordinates, in degrees.
    dec_column : str
        The name of the column in ``table`` that contains the Declination
        coordinates, in degrees.
    mag_column : str
        The name of the column in ``table`` with the magnitude to be used to
        scale ``min_separation``.
    is_flux : bool
        If `True`, assumes the ``mag_column`` values are given as fluxes
        (units of flux_unit).
    flux_unit : str
        Gives the units of flux in the 'mag_column' - known values 'nMgy', 'Jy'
    mag_threshold : float
        The value below which the separation to neighbouring sources will be
        scaled.
    radius_column : str
         Name of the database column that provided the object radius
         (an alternative to mag_column useful for extended sources)
    scale_a : float
         Value of :math:`a` in the radius vs mag relationship
    scale_b : float
         Value of :math:`\beta` in the radius vs mag relationship
    calculate_min_separation : bool
        If `True`, calculates the separation to the nearest neighbour for
        each candidate sky position.
    nsample : int or None
        The total number of skies to retrieve for each tile. If `None`,
        returns all candidate skies.
    downsample_nside : int
        The HEALPix nside used for downsampling. If ``nside``, the
        resulting valid skies will be grouped by HEALPix pixels of this
        resolution. For each pixel a random sample will be drawn so that the
        total number of skies selected matches ``nsample``.
    downsample_data : pandas.DataFrame
        A data frame with previously selected skies that will be used to
        downsample the sky candidates. This is useful when trying to create a
        sky catalogue from multiple tables to ensure that the selected sky
        positions match across the various tables. If not enough valid skies
        can be selected from the sample in the data frame, they will be
        completed up to ``nsample``.
    n_cpus : int
        Number of CPUs to use for multiprocessing.
    seed : int
        The random state seed.

    Returns
    -------
    skies : pandas.DataFrame
        The list of selected skies.

    """

    assert database.connected, "database is not connected."

    columns = (
        f"healpix_ang2ipix_nest("
        f"{tile_nside}, {ra_column}, {dec_column}) "
        f"AS tile_{tile_nside}, "
        f"{ra_column} AS ra, {dec_column} AS dec"
    )

    if mag_column and mag_threshold:
        columns += f", {mag_column} AS mag"

    if radius_column:
        columns += f", {radius_column} as radius"

    if tiles is None:
        tiles = numpy.arange(healpy.nside2npix(tile_nside))

    if downsample_data is not None:
        downsample_data = downsample_data.loc[:, ["tile_32", "valid"]]
        downsample_data = downsample_data.loc[downsample_data.valid]

    query = (
        f"SELECT {columns} FROM {table} "
        f"WHERE healpix_ang2ipix_nest("
        f"{tile_nside}, {ra_column}, {dec_column}) = {{tile}};"
    )

    pbar = enlighten.Counter(total=len(tiles), desc=output, unit="tiles")
    pbar.refresh()

    database_params = database.connection_params
    database_params.update({"database": database.dbname})

    process_tile = partial(
        _process_tile,
        database_params=database_params,
        query=query,
        candidate_nside=candidate_nside,
        tile_nside=tile_nside,
        min_separation=min_separation,
        is_flux=is_flux,
        flux_unit=flux_unit,
        scale_a=scale_a,
        scale_b=scale_b,
        mag_threshold=mag_threshold,
        nsample=nsample,
        downsample_data=downsample_data,
        calculate_min_separation=calculate_min_separation,
        downsample_nside=downsample_nside,
        seed=seed,
    )

    all_skies: pandas.DataFrame | None = None

    with multiprocessing.Pool(n_cpus) as pool:
        for tile_skies in pool.imap_unordered(process_tile, tiles, chunksize=5):
            if tile_skies is not False and len(tile_skies) > 0:
                if all_skies is None:
                    all_skies = tile_skies
                else:
                    all_skies = all_skies.append(tile_skies)

            pbar.update()

    if all_skies is None:
        return all_skies

    # Not sure why this is needed but it seems downsampling sometimes
    # converts the "valid" column to object.
    all_skies.valid = all_skies.valid.astype(bool)

    # Downcast some columns
    all_skies = all_skies.astype({f"tile_{tile_nside}": numpy.int16})

    if "sep_neighbour" in all_skies:
        all_skies = all_skies.astype({"sep_neighbour": numpy.float32})
    if "mag_neighbour" in all_skies:
        all_skies = all_skies.astype({"mag_neighbour": numpy.float32})

    if downsample_nside is not None:
        all_skies = all_skies.astype({"down_pix": numpy.int32})

    all_skies.to_hdf(output, key="data")

    return all_skies


def plot_sky_density(file_or_data, nside, pix_column=None, nside_plot=32, **kwargs):
    """Plots the number of skies as a HEALPix map.

    Parameters
    ----------
    file_or_data : str or ~pandas.DataFrame
        A HDF5 file with the sky catalogue or a Pandas data frame with the
        data. It is assumed that the file or data frame contains at least a
        column ``pix_X``, where ``X`` is the order corresponding to ``nside``.
    nside : int
        The HEALPix nside for the skies.
    pix_column : str or None
        The column that contains the HEALPix pixels. If `None`, the index of
        the data frame will be used.
    nside_plot : int
        The HEALPix nside in which the data will be plotted.
    kwargs : dict
        Other keyword parameters to pass to healpy's
        `~healpy.visufunc.mollview`.

    Returns
    -------
    figure : `~matplotlib.figure.Figure`
        The Matplotlib figure with the plot.

    """

    if isinstance(file_or_data, str):
        data = pandas.read_hdf(file_or_data)
    else:
        data = file_or_data.copy()

    if not pix_column:
        pix_column = data.index.name
        data.reset_index(inplace=True)

    data[f"pix_{nside_plot}"] = nested_regrade(data[pix_column], nside, nside_plot)
    count = data.groupby(f"pix_{nside_plot}").count()

    hmap = numpy.arange(healpy.nside2npix(nside_plot), dtype=numpy.float32)
    hmap[:] = healpy.UNSEEN
    hmap[count.index] = count[pix_column]

    figure = plt.figure()
    healpy.mollview(hmap, fig=figure.number, nest=True, **kwargs)

    return figure


def plot_skies(
    file_or_data,
    ra,
    dec,
    radius=1.5,
    targets=None,
    show_sky_buffer=False,
    buffer_radius=10.0,
):
    """Plots the skies (and optionally targets) in a regions.

    Parameters
    ----------
    file_or_data : str or ~pandas.DataFrame
        A HDF5 file with the sky catalogue or a Pandas data frame with the
        data, including columns labelled ``ra`` and ``dec``.
    ra : float
        The right ascension of the centre of the field.
    dec : float
        The declination of the centre of the field.
    radius : float
        The FOV radius, in degrees.
    targets : list
        A target list or array as ``(ra, dec)``.
    show_sky_buffer : bool
        Plots the buffer regions around each sky.
    buffer_radius : float
        The sky buffer radius, in arcsec.

    Returns
    -------
    figure : `~matplotlib.figure.Figure`
        The Matplotlib figure with the plot.

    """

    if isinstance(file_or_data, str):
        data = pandas.read_hdf(file_or_data)
    else:
        data = file_or_data

    fig, ax = plt.subplots()

    cos_factor = numpy.cos(numpy.radians(dec))

    FOV = Ellipse(
        (ra, dec),
        radius * 2 / cos_factor,
        radius * 2,
        linewidth=2,
        color="None",
        ec="k",
    )
    ax.add_patch(FOV)

    # Plot skies
    centre = SkyCoord(ra=ra, dec=dec, unit="deg")
    sky_coords = SkyCoord(ra=data.ra.to_numpy(), dec=data.dec.to_numpy(), unit="deg")
    skies = data[sky_coords.separation(centre).value < radius]

    ax.scatter(skies.ra, skies.dec, marker="o", s=0.2, color="b", zorder=10, label="Skies")

    if show_sky_buffer:
        for _, sky in skies.iterrows():
            buffer = Ellipse(
                (sky.ra, sky.dec),
                buffer_radius / 3600.0 * 2,
                buffer_radius / 3600.0 * 2,
                color="y",
                ec="None",
                alpha=0.2,
                zorder=0,
            )
            ax.add_patch(buffer)

    # Plot targets
    if targets is not None:
        targets = numpy.array(targets)
        target_coords = SkyCoord(ra=targets[:, 0], dec=targets[:, 1], unit="deg")
        targets = targets[target_coords.separation(centre).value < radius]

        ax.scatter(
            targets[:, 0],
            targets[:, 1],
            marker="x",
            s=0.2,
            color="r",
            zorder=20,
            label="Targets",
        )

    ax.legend(loc="upper right")

    ax.set_xlim(ra - (radius - 0.1) / cos_factor, ra + (radius + 0.1) / cos_factor)
    ax.set_ylim(dec - radius - 0.1, dec + radius + 0.1)

    ax.set_xlabel("Right Ascension [deg]")
    ax.set_ylabel("Declination [deg]")

    ax.set_aspect(1 / cos_factor)

    fig.tight_layout()

    return fig


# This is the main function to use with catalogdb as catalog
def create_sky_catalogue(database, tiles=None, **kwargs):
    """A script to generate a combined sky catalogue from multiple sources."""

    default_mag_threshold = 14.0
    default_min_separation = 5.0
    default_param_a = 0.15
    default_param_b = 1.5
    nsample = 2048

    if not os.path.exists("tmass_skies.h5"):
        log.info("Procesing twomass_psc.")
        get_sky_table(
            database,
            "catalogdb.twomass_psc",
            "tmass_skies.h5",
            dec_column="decl",
            mag_column="h_m",
            mag_threshold=default_mag_threshold,
            min_separation=default_min_separation,
            scale_a=default_param_a,
            scale_b=default_param_b,
            nsample=nsample,
            tiles=tiles,
            **kwargs,
        )
    else:
        warnings.warn("Found file tmass_skies.h5", TargetSelectionUserWarning)

    # We use 2MASS as the source for the downsampled candidates.
    tmass = pandas.read_hdf("tmass_skies.h5")

    if not os.path.exists("gaia_skies.h5"):
        log.info("Procesing gaia_dr2_source.")
        get_sky_table(
            database,
            "catalogdb.gaia_dr2_source",
            "gaia_skies.h5",
            mag_column="phot_g_mean_mag",
            mag_threshold=default_mag_threshold,
            min_separation=default_min_separation,
            scale_a=default_param_a,
            scale_b=default_param_b,
            nsample=nsample,
            downsample_data=tmass,
            tiles=tiles,
            **kwargs,
        )
    else:
        warnings.warn("Found file gaia_skies.h5", TargetSelectionUserWarning)

    if not os.path.exists("ps1dr2_skies.h5"):
        log.info("Procesing ps1dr2.")
        get_sky_table(
            database,
            "catalogdb.panstarrs1",
            "ps1dr2_skies.h5",
            mag_column="r_stk_psf_flux",
            is_flux=True,
            flux_unit="Jy",
            mag_threshold=default_mag_threshold,
            min_separation=default_min_separation,
            scale_a=default_param_a,
            scale_b=default_param_b,
            nsample=nsample,
            downsample_data=tmass,
            tiles=tiles,
            **kwargs,
        )
    else:
        warnings.warn("Found file ps1dr2_skies.h5", TargetSelectionUserWarning)

    if not os.path.exists("ls8_skies.h5"):
        log.info("Procesing legacy_survey_dr8.")
        get_sky_table(
            database,
            "catalogdb.legacy_survey_dr8",
            "ls8_skies.h5",
            mag_column="flux_r",
            is_flux=True,
            flux_unit="nMgy",
            mag_threshold=default_mag_threshold,
            min_separation=default_min_separation,
            scale_a=default_param_a,
            scale_b=default_param_b,
            nsample=nsample,
            downsample_data=tmass,
            tiles=tiles,
            **kwargs,
        )
    else:
        warnings.warn("Found file ls8_skies.h5", TargetSelectionUserWarning)

    if not os.path.exists("tycho2_skies.h5"):
        log.info("Procesing tycho2.")
        get_sky_table(
            database,
            "catalogdb.tycho2",
            "tycho2_skies.h5",
            ra_column="ramdeg",
            dec_column="demdeg",
            mag_column="vtmag",
            mag_threshold=default_mag_threshold,
            min_separation=default_min_separation,
            scale_a=default_param_a,
            scale_b=default_param_b,
            nsample=nsample,
            downsample_data=tmass,
            tiles=tiles,
            **kwargs,
        )
    else:
        warnings.warn("Found file tycho2_skies.h5", TargetSelectionUserWarning)

    if not os.path.exists("tmass_xsc_skies.h5"):
        log.info("Procesing twomass_xsc.")
        get_sky_table(
            database,
            "catalogdb.twomass_xsc",
            "tmass_xsc_skies.h5",
            dec_column="decl",
            radius_column="r_ext",
            # mag_column='h_m_k20fe', mag_threshold=14,
            nsample=nsample,
            downsample_data=tmass,
            tiles=tiles,
            **kwargs,
        )
    else:
        warnings.warn("Found file tmass_xsc_skies.h5", TargetSelectionUserWarning)

    skies: pandas.DataFrame | None = None
    col_order = []

    for file_ in [
        "gaia_skies.h5",
        "ls8_skies.h5",
        "ps1dr2_skies.h5",
        "tmass_skies.h5",
        "tycho2_skies.h5",
        "tmass_xsc_skies.h5",
    ]:
        table_name = file_[0:-9]

        if not os.path.exists(file_):
            warnings.warn(f"File {file_} does not exists.", TargetSelectionUserWarning)
            continue

        table = pandas.read_hdf(file_).drop_duplicates()
        assert isinstance(table, pandas.DataFrame)

        table.rename(
            columns={
                "sep_neighbour": f"sep_neighbour_{table_name}",
                "mag_neighbour": f"mag_neighbour_{table_name}",
                "valid": f"valid_{table_name}",
                "selected": f"selected_{table_name}",
            },
            inplace=True,
        )

        if skies is None:
            skies = table
        else:
            skies = skies.combine_first(table)

        for col in [
            f"valid_{table_name}",
            f"selected_{table_name}",
            f"sep_neighbour_{table_name}",
            f"mag_neighbour_{table_name}",
        ]:
            if col in skies:
                col_order.append(col)

    assert skies is not None
    skies = skies.loc[:, ["ra", "dec", "down_pix", "tile_32"] + col_order]

    for col in skies:
        if str(col).startswith("valid_") or str(col).startswith("selected_"):
            skies[col].fillna(False, inplace=True)

    # Now do some masking based on healpixels that are flagged by a
    # conservative veto_mask. This is intended to catch a few cases
    # where the nearest neighbour star is not the one that is
    # contributing the most flux at a candidate sky location

    # Now get the list of all healpixels that are within circles around
    # very bright stars:
    # veto_mask = create_veto_mask(database=database, nside=32768,
    #                              moc_filename='veto_from_tycho2.moc',
    #                              table='catalogdb.tycho2',
    #                              ra_column='ramdeg', dec_column='demdeg',
    #                              mag_column='vtmag', mag_threshold=12,
    #                              min_separation=15.0, param_a=0.15,
    #                              param_b=1.5, overwrite=True)

    # # Now flag any skies that have a 'pix_32768' that is in the veto_mask
    # skies.loc[:, 'tycho2_veto'] = False
    # skies.loc[skies.index.isin(veto_mask), 'tycho2_veto'] = True

    skies.tile_32 = skies.tile_32.astype(int)
    skies.down_pix = skies.down_pix.astype(int)
    skies.to_hdf("skies.h5", key="data")


def create_veto_mask(
    database,
    nside=32768,
    moc_filename=None,
    overwrite=False,
    debug_limit=None,
    table="gaia_dr2_source",
    ra_column="ra",
    dec_column="dec",
    mag_column="phot_g_mean_mag",
    mag_threshold=12.0,
    min_separation=15.0,
    param_a=0.15,
    param_b=1.5,
):
    """Generate a list of healpixels that must be avoided because
    they lie within the near-zones of bright stars (and/or galaxies TBD).

    I have a hunch that generating this list once and then testing skies
    against it will be more efficient (and reliable) than checking candidate
    skies against a list of stars. This avoids potential pitfalls of nearest
    neighbour method (e.g. when second nearest neighbour is brighter than
    nearest neighbour).

    This doesn't take too long to run, so result does not need to be preserved
    long term. However, writing out a MOC file is a convenient way to visualise
    what has been done.

    Parameters
    ----------
    database : ~sdssdb.connection.PeeweeDatabaseConnection
        A valid database connection.
    nside : int
        HEALPix resolution of the returned healpix pixel list
    moc_filename : str
        Path to the MOC file to write (or None).
    overwrite: bool
        Whether to clobber the MOC file
    debug_limit: int
        Max number of stars to return in database query - debug purposes only
    ra_column : str
        The name of the column in ``table`` that contains the Right Ascension
        coordinates, in degrees.
    dec_column : str
        The name of the column in ``table`` that contains the Declination
        coordinates, in degrees.
    mag_column : str
        The name of the column in ``table`` with the magnitude to be used to
        scale ``min_separation``.
    param_a: float
        A parameter that controls the how the radius scales with magnitude
    param_b: float
        A parameter that controls the how the radius scales with magnitude

    Returns
    -------
    mask : ~numpy.ndarray
        A (numpy) array of healpixel indices (resolution ``nside``) that
        fall within the mask.

    """

    as2rad = numpy.pi / (180.0 * 3600.0)
    hpx_order = healpy.nside2order(nside)
    pixarea = healpy.nside2pixarea(nside)

    # get the list of gaia dr2 stars brighter than G=12 from the database

    query = (
        f"SELECT {ra_column},{dec_column},{mag_column} from "
        f"{table} WHERE {mag_column} < {mag_threshold} AND "
        f"{ra_column} IS NOT NULL AND {dec_column} IS NOT NULL"
    )
    if debug_limit is not None:
        query = query + f"limit {debug_limit} "

    targets = pandas.read_sql(query, database)
    print(
        f"Working on {len(targets):,} bright stars "
        f"({mag_column} < {mag_threshold}) from {table}"
    )

    # compute coords on unit sphere
    vector = healpy.pixelfunc.ang2vec(targets[ra_column], targets[dec_column], lonlat=True)

    # compute mag-dependent exclusion radii
    corr = numpy.power(mag_threshold - targets[mag_column], param_b) / param_a
    radius = as2rad * min_separation + corr

    ipix_list = []
    for v, r in zip(vector, radius):
        i = healpy.query_disc(nside, vec=v, radius=r, inclusive=True, fact=4, nest=True)
        if len(i) > 0:
            ipix_list.extend(list(i))

    # we only one copy of each masked pixel:
    ipix = numpy.unique(ipix_list)
    npix = len(ipix)
    print(f"Result: {npix:,} masked pixels (NSIDE={nside}), area={npix*pixarea:.4f} sqdeg")

    if moc_filename is not None:
        m = MOC.from_healpix_cells(ipix=ipix, depth=numpy.repeat(hpx_order, len(ipix)))
        m.write(moc_filename, format="fits", overwrite=overwrite)

    return ipix


@overload
def is_valid_sky(
    coords: npt.NDArray[numpy.float_] | Sequence[Sequence[float]],
    database: PeeweeDatabaseConnection | str,
    catalogues: list | None = None,
    param_overrides: dict | None = None,
    epoch: float | None = None,
    fibre_radius: float | Literal["APO", "LCO"] = 1.0,
    return_dataframe: Literal[False] = False,
) -> npt.NDArray[numpy.bool_]: ...


@overload
def is_valid_sky(
    coords: npt.NDArray[numpy.float_] | Sequence[Sequence[float]],
    database: PeeweeDatabaseConnection | str,
    catalogues: list | None = None,
    param_overrides: dict | None = None,
    epoch: float | None = None,
    fibre_radius: float | Literal["APO", "LCO"] = 1.0,
    return_dataframe: Literal[True] = True,
) -> tuple[npt.NDArray[numpy.bool_], polars.DataFrame]: ...


def is_valid_sky(
    coords: npt.NDArray[numpy.float_] | Sequence[Sequence[float]],
    database: PeeweeDatabaseConnection | str,
    catalogues: list | None = None,
    param_overrides: dict | None = None,
    epoch: float | None = None,
    fibre_radius: float | Literal["APO", "LCO"] = 1.0,
    return_dataframe: bool = False,
) -> npt.NDArray[numpy.bool_] | tuple[npt.NDArray[numpy.bool_], polars.DataFrame]:
    """Check if a set of coordinates are valid sky positions.

    Parameters
    ----------
    coords
        An ``Nx2`` array with the RA/Dec coordinates to check.
    database
        A valid database connection, either an ``sdssdb.connection.PeeweeDatabaseConnection``
        or a valid Postgresql connection string.
    catalogues
        A list of catalogues to check the positions against. Must be valid tables in
        ``sdss5db.catalogdb``. Defaults to the keys in `.DEFAULT_CATALOGUE_PARAMS`.
    param_overrides
        A dictionary with overrides for the catalogue parameters. See `.DEFAULT_CATALOGUE_PARAMS`
        for a list of the default parameters. The parameters defined here are merged on top of
        `.DEFAULT_CATALOGUE_PARAMS` and used to query the database.
    epoch
        The epoch of the coordinates as a Julian date. If `None`, the current epoch is assumed.
        This is only used to propagate catalogue coordinates with proper motions to the same
        epoch as the candidate coordinates.
    fibre_radius
        The radius of the fibre in arcsec. If 'APO' or 'LCO', the default values for the
        respective telescopes are used. Any candidate coordinate that is within this radius
        of a catalogue target is marked as not a sky.
    return_dataframe
        If `True`, returns a dataframe with the results.

    Returns
    -------
    valid
        A boolean array with the same length as the input coordinates indicating if the
        positions are valid sky regions.

    """

    # Sanity checks.
    coords = numpy.atleast_2d(coords)
    assert len(coords.shape) == 2 and coords.shape[1] == 2, "coords must be a Nx2 array."

    catalogues = catalogues or list(DEFAULT_CATALOGUE_PARAMS)

    if not isinstance(database, str):
        assert database.connected, "database is not connected."

        for cat_name in catalogues:
            assert database.table_exists(cat_name, "catalogdb"), f"Table {cat_name} not found."

    if fibre_radius == "APO":
        fibre_radius = 1.0
    elif fibre_radius == "LCO":
        fibre_radius = 0.654

    if epoch is None:
        epoch_ap = Time.now()
    else:
        epoch_ap = Time(epoch, format="jd")

    # Create the output dataframe.
    df = (
        polars.from_numpy(
            coords,
            schema={"ra": polars.Float64, "dec": polars.Float64},
            orient="row",
        )
        .with_columns(**{cat: polars.lit(True, polars.Boolean) for cat in catalogues})
        .with_row_count("n")
    )

    # Get and validate catalogue parameters.
    c_params: dict[str, CatalogueParams] = {}
    for cat_name in catalogues:
        if cat_name in DEFAULT_CATALOGUE_PARAMS:
            default_params = DEFAULT_CATALOGUE_PARAMS[cat_name].copy()
        else:
            if param_overrides is None or cat_name not in param_overrides:
                raise ValueError(
                    f"Catalogue {cat_name} has no defaults and was not found in param_overrides."
                )
            default_params = {}

        # Update the default params with the overrides.
        default_params.update((param_overrides or {}).get(cat_name, {}))

        c_params[cat_name] = CatalogueParams(**default_params)

    # Calculate the centre and radius of the region to query to include all the input positions.
    ra_cen = circmean(coords[:, 0], 360)
    dec_cen = coords[:, 1].mean()
    cen = SkyCoord(ra=ra_cen, dec=dec_cen, unit="deg")
    coords_ap = SkyCoord(ra=coords[:, 0], dec=coords[:, 1], unit="deg")
    sep = coords_ap.separation(cen)
    radius = sep.max().deg + 0.1  # Add a small buffer to account for edge effects.

    if radius > 3:
        raise ValueError(
            "The radius of the region to query is too large. "
            "Please provide coordinates that span a smaller region."
        )

    # Iterate over each catalogue and query sources in the region.
    for cat_name, cat_params in c_params.items():
        radius_select = ""
        if cat_params.radius_column is not None:
            radius_select = f"{cat_params.radius_column} AS radius"

        mag_or_flux_select = ""
        if cat_params.mag_column is not None:
            mag_or_flux_select = f"{cat_params.mag_column} AS mag_or_flux"

        pm_selects = []
        has_pm: bool = False
        if cat_params.pmra_column:
            pm_selects.append(f"{cat_params.pmra_column} AS pmra")
            pm_selects.append(f"{cat_params.pmdec_column} AS pmdec")
            pm_selects.append(f"{cat_params.ref_epoch} AS ref_epoch")
            has_pm = True

        selects = [
            f"{cat_params.ra_column} AS ra",
            f"{cat_params.dec_column} AS dec",
            mag_or_flux_select,
            radius_select,
        ]
        selects += pm_selects

        select = ", ".join([sel for sel in selects if sel != ""])
        query = (
            f"SELECT {select} "
            f"FROM catalogdb.{cat_name} "
            f"WHERE q3c_radial_query({cat_params.ra_column}, {cat_params.dec_column}, "
            f"{ra_cen}, {dec_cen}, {radius})"
        )

        # Get all the targets in the query region.
        if isinstance(database, str):
            q_targets = polars.read_database_uri(query, database, engine="adbc")
        else:
            q_targets = polars.read_database(query, database)

        q_targets = q_targets.with_row_count(name="n")

        if cat_params.ref_epoch is not None:
            q_targets = q_targets.cast({"ref_epoch": polars.Float32})

        if len(q_targets) == 0:
            continue

        # Check which of our candidate coordinates would be within a fibre radius of any
        # of the catalogue targets. Mark those as NOT skies regardless of the magnitude.
        ra = q_targets["ra"].to_numpy()
        dec = q_targets["dec"].to_numpy()
        if has_pm:
            epoch_delta = epoch_ap.jyear - q_targets["ref_epoch"].to_numpy()

            pmra_deg = q_targets["pmra"].to_numpy() / 3600.0 / 1000.0
            pmdec_deg = q_targets["pmdec"].to_numpy() / 3600.0 / 1000.0
            pmra_deg[numpy.isnan(pmra_deg)] = 0.0
            pmdec_deg[numpy.isnan(pmdec_deg)] = 0.0

            ra = ra + pmra_deg * epoch_delta / numpy.cos(numpy.radians(dec))
            dec = dec + pmdec_deg * epoch_delta

        q_targets_ap = SkyCoord(ra=ra, dec=dec, unit="deg")

        _, sep2d, _ = match_coordinates_sky(coords_ap, q_targets_ap)
        fibre_mask = sep2d.arcsec > fibre_radius  # True if nothing found within fibre_radius
        df = df.with_columns(polars.Series(fibre_mask, dtype=polars.Boolean).alias(cat_name))

        # Now that we have checked the closest neighbour, keep only the "bright" targets
        # by applying the magniture threshold, and calculate the target magnitude and separation.
        if cat_params.mag_column is not None:
            if cat_params.is_flux:
                zpt = _known_flux_zpts[cat_params.flux_unit]
                b_targets = q_targets.with_columns(mag=zpt - 2.5 * polars.col.mag_or_flux.log10())
            else:
                b_targets = q_targets.with_columns(mag=polars.col.mag_or_flux)
            b_targets = b_targets.filter(polars.col.mag_or_flux <= cat_params.mag_threshold)
        else:
            b_targets = q_targets

        if cat_params.radius_column is not None:
            b_targets = b_targets.with_columns(min_sep=cat_params.min_separation)
            b_targets = b_targets.with_columns(min_sep=polars.max_horizontal("radius", "min_sep"))
        else:
            b_targets = b_targets.with_columns(
                min_sep=(
                    cat_params.min_separation
                    + (cat_params.mag_threshold - polars.col.mag).pow(cat_params.scale_b)
                    / cat_params.scale_a
                )
            )

        b_targets_ap = q_targets_ap[b_targets["n"].to_numpy()]

        # For each bright target, get all the input coordinate matches that are within the
        # maximum minimum separation (search_around_sky only accepts a scalar separation).
        idx1, idx2, sep2d, _ = search_around_sky(
            b_targets_ap,
            coords_ap,
            b_targets["min_sep"].to_numpy().max() * uu.arcsec,
        )

        # Get the real minimum separation for the matched bright targets.
        min_sep = b_targets[idx1]["min_sep"].to_numpy()

        # Get the matches that are really within the minimum separation per bright target.
        idx2_min_sep = idx2[sep2d.arcsec < min_sep]
        idx2_min_sep.sort()

        # Mark the entries in that are within minimum separation as NOT skies.
        df[idx2_min_sep, cat_name] = False

    # Create the valid sky mask by combining all the catalogues.
    valid_df = df.select(*catalogues)
    mask = valid_df.fold(lambda s1, s2: s1 & s2).to_numpy().astype(numpy.bool_)

    if return_dataframe:
        return (mask, df)

    return mask



# encoding: utf-8

import os

import enlighten

from sdsstools import get_config, get_logger, get_package_version


NAME = "sdss-target-selection"

config = get_config(
    "target_selection",
    os.path.join(os.path.dirname(__file__), "config/target_selection.yml"),
)

log = get_logger(NAME)


__version__ = get_package_version(path=__file__, package_name=NAME)


manager = enlighten.get_manager()

# Avoid an annoying message when importing dustmaps
with open(os.path.expanduser("~/.dustmapsrc"), "w") as f:
    f.write("{}\n")

from .xmatch import XMatchPlanner, XMatchModel  # isort:skip



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2023-04-24
# @Filename: mwm_monitor_apogee.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    SDSS_DR17_APOGEE_Allstarmerge,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# peewee model : postgres table name
# Catalog : catalogdb.catalog
# Gaia_DR3 : catalogdb.gaia_dr3_source
# SDSS_DR17_APOGEE_Allstarmerge : catalogdb.sdss_dr17_apogee_allstarmerge
# TwoMassPSC : catalogdb.twomass_psc
#
#
# We associate each APOGEE target with a 2MASS source since the
# vast majority of APOGEE targets were selected from the 2MASS PSC.
# APOGEE_ID is the same as
# the “designation” column of the 2MASS PSC;
#
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# For sdss_dr17_apogee_allstarmerge we
# have to strip off the “2M” in the APOGEE_ID.
# However, some apogee_id values do not have 2M in the front.
# (old: for sdss_apogeeallstarmerge_r13 we also had to strip off the
#  “2M” in the APOGEE_ID.)
# For example:
# sdss5db=# select designation from  catalogdb.twomass_psc limit 2;
#    designation
# ------------------
#  12034281-5740002
#  12032366-5738380
# (2 rows)
#
# sdss5db=# select apogee_id from catalogdb.SDSS_DR17_APOGEE_Allstarmerge
# where apogee_id like '2M%';
#      apogee_id
# ---------------------
# 2M00000002+7417074
# 2M00000019-1924498
# etc.
# sdss5db=# select apogee_id from catalogdb.SDSS_DR17_APOGEE_Allstarmerge
# where apogee_id not like '2M%';
#    apogee_id
# --------------------
# 19140272-1554055
# 19155129-1617591
#
#


class MWM_monitor_apogee_Base_Carton(BaseCarton):
    """
    MWM_monitor_apogee_Base_Carton is a base carton.

    Actual cartons are implemented as subclasses of MWM_monitor_apogee_Base_Carton
    for different conditions in the WHERE clause.

    5.1.25. mwm_monitor_*apogee_*

    Shorthand name: mwm_monitor_apogee

    Existing carton code: None

    Simplified Description of selection criteria:
    These are targets selected from globular and open clusters that
    have been observed many times over many years with the apogee instrument.

    mwm_monitor_n188_apogee_long:
    SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%N188%'
    AND baseline >= 3000 AND nvisits >= 12  high priority

    mwm_monitor_n188_apogee_short:
    SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%N188%'
    AND baseline < 3000 AND baseline >= 1800 AND nvisits >= 12  lower priority

    mwm_monitor_m67_apogee_long:
    SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%M67%'
    AND baseline >= 2000 AND nvisits >= 12  high priority

    mwm_monitor_m67_apogee_short:
    SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%M67%'
    AND baseline < 2000 AND baseline >= 1300 AND nvisits >= 12  lower priority

    mwm_monitor_m15_apogee_long:
    SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%M15%'
    AND baseline >= 1300 AND nvisits >= 12  high priority

    mwm_monitor_m15_apogee_short:
    SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%M15%'
    AND baseline < 1300 AND baseline >= 900 AND nvisits >= 12  lower priority

    Gaia DR2 parameters to be converted to Gaia DR3: Nothing to be converted.

    Return columns: RA, DEC, proper motion, hmag, 2mass name.

    Metadata:

    Priority:
    High priority=1300
    Lower Priority =1302

    Cadence: bright_1x4

    Instrument: APOGEE

    can_offset=True

    Lead contact: Nathan De Lee
    """

    # SELECT apogee_id,ra,dec,nvisits,baseline,j,j_err,h,h_err,k,k_err,fields
    # FROM catalogdb.sdss_dr17_apogee_allstarmerge WHERE fields LIKE '%N188%'
    # AND baseline >= 3000 AND nvisits >= 12  high priority

    # In the below query, we use replace() instead of ltrim() since
    # ltrim('2M20', '2M') will also trim the second 2.

    def build_query(self, version_id, query_region=None):
        query = (
            Catalog.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag,
                TwoMassPSC.pts_key,
                TwoMassPSC.designation,
                SDSS_DR17_APOGEE_Allstarmerge.apogee_id,
                SDSS_DR17_APOGEE_Allstarmerge.ra,
                SDSS_DR17_APOGEE_Allstarmerge.dec,
                SDSS_DR17_APOGEE_Allstarmerge.nvisits,
                SDSS_DR17_APOGEE_Allstarmerge.baseline,
                SDSS_DR17_APOGEE_Allstarmerge.j,
                SDSS_DR17_APOGEE_Allstarmerge.j_err,
                SDSS_DR17_APOGEE_Allstarmerge.h,
                SDSS_DR17_APOGEE_Allstarmerge.h_err,
                SDSS_DR17_APOGEE_Allstarmerge.k,
                SDSS_DR17_APOGEE_Allstarmerge.k_err,
                SDSS_DR17_APOGEE_Allstarmerge.fields,
            )
            .join(CatalogToGaia_DR3, on=(Catalog.catalogid == CatalogToGaia_DR3.catalogid))
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .join(
                SDSS_DR17_APOGEE_Allstarmerge,
                on=(
                    TwoMassPSC.designation
                    == peewee.fn.replace(SDSS_DR17_APOGEE_Allstarmerge.apogee_id, "2M", "")
                ),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """
        Select the brightest sources (i.e. smallest phot_g_mean_mag)
        for duplicate apogee_id (but different gaia dr3 source id)
        """

        # self.name is the carton name
        temp_table_name = "sandbox.temp_" + self.name
        self.database.execute_sql("update " + temp_table_name + " set selected = false")

        cursor = self.database.execute_sql(
            "select catalogid, apogee_id, phot_g_mean_mag from "
            + temp_table_name
            + " order by apogee_id asc, phot_g_mean_mag asc;"
        )

        output = cursor.fetchall()
        list_of_catalog_id = [0] * len(output)

        # We use the name counter instead of count
        # since count is a python built-in function.
        # For counter we use a dictionary
        # since current_apogee_id is a string.
        counter = {}

        current_target = 0
        for i in range(len(output)):
            current_apogee_id = output[i][1]
            # If two targets have same apogee id then we only keep the
            # first source. Since the sources are
            # ordered by apogee_id (ascending) and phot_g_mean_mag (ascending)
            # this means that we will keep the sources with
            # smallest phot_g_mean_mag (i.e. brightest sources).
            if current_apogee_id not in counter.keys():
                counter[current_apogee_id] = True
                list_of_catalog_id[current_target] = output[i][0]
                current_target = current_target + 1

        max_target = current_target
        for k in range(max_target + 1):
            self.database.execute_sql(
                " update "
                + temp_table_name
                + " set selected = true "
                + " where catalogid = "
                + str(list_of_catalog_id[k])
                + ";"
            )


# Below are sub classes of the above base class MWM_monitor_apogee_Base_Carton.
# Note that there is no __init__() in below sub classes since
# they use the __init__() from the BaseCarton class in base.py.


class MWM_monitor_n188_apogee_long_Carton(MWM_monitor_apogee_Base_Carton):
    name = "mwm_monitor_n188_apogee_long"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x4"
    program = "mwm_monitor"
    mapper = "MWM"
    priority = 1300
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            SDSS_DR17_APOGEE_Allstarmerge.baseline >= 3000,
            SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 12,
            SDSS_DR17_APOGEE_Allstarmerge.fields.contains("N188"),
        )

        return query

    def post_process(self, model):
        # Note that self is a parameter of post_process() above but
        # self is not a parameter of post_process() below
        # since we are using super().
        super().post_process(model)


class MWM_monitor_n188_apogee_short_Carton(MWM_monitor_apogee_Base_Carton):
    name = "mwm_monitor_n188_apogee_short"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x4"
    program = "mwm_monitor"
    mapper = "MWM"
    priority = 1302
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            SDSS_DR17_APOGEE_Allstarmerge.baseline >= 1800,
            SDSS_DR17_APOGEE_Allstarmerge.baseline < 3000,
            SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 12,
            SDSS_DR17_APOGEE_Allstarmerge.fields.contains("N188"),
        )

        return query

    def post_process(self, model):
        super().post_process(model)


class MWM_monitor_m67_apogee_long_Carton(MWM_monitor_apogee_Base_Carton):
    name = "mwm_monitor_m67_apogee_long"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x4"
    program = "mwm_monitor"
    mapper = "MWM"
    priority = 1300
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            SDSS_DR17_APOGEE_Allstarmerge.baseline >= 2000,
            SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 12,
            SDSS_DR17_APOGEE_Allstarmerge.fields.contains("M67"),
        )

        return query

    def post_process(self, model):
        super().post_process(model)


class MWM_monitor_m67_apogee_short_Carton(MWM_monitor_apogee_Base_Carton):
    name = "mwm_monitor_m67_apogee_short"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x4"
    program = "mwm_monitor"
    mapper = "MWM"
    priority = 1302
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            SDSS_DR17_APOGEE_Allstarmerge.baseline >= 1300,
            SDSS_DR17_APOGEE_Allstarmerge.baseline < 2000,
            SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 12,
            SDSS_DR17_APOGEE_Allstarmerge.fields.contains("M67"),
        )

        return query

    def post_process(self, model):
        super().post_process(model)


class MWM_monitor_m15_apogee_long_Carton(MWM_monitor_apogee_Base_Carton):
    name = "mwm_monitor_m15_apogee_long"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x4"
    program = "mwm_monitor"
    mapper = "MWM"
    priority = 1300
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            SDSS_DR17_APOGEE_Allstarmerge.baseline >= 1300,
            SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 12,
            SDSS_DR17_APOGEE_Allstarmerge.fields.contains("M15"),
        )

        return query

    def post_process(self, model):
        super().post_process(model)


class MWM_monitor_m15_apogee_short_Carton(MWM_monitor_apogee_Base_Carton):
    name = "mwm_monitor_m15_apogee_short"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x4"
    program = "mwm_monitor"
    mapper = "MWM"
    priority = 1302
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        query = query.where(
            SDSS_DR17_APOGEE_Allstarmerge.baseline >= 900,
            SDSS_DR17_APOGEE_Allstarmerge.baseline < 1300,
            SDSS_DR17_APOGEE_Allstarmerge.nvisits >= 12,
            SDSS_DR17_APOGEE_Allstarmerge.fields.contains("M15"),
        )

        return query

    def post_process(self, model):
        super().post_process(model)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-04-28
# @Filename: mwm_magcloud_agb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import pandas
import peewee
from astropy.coordinates import SkyCoord
from astropy.units import mas, yr
from gala.coordinates import MagellanicStreamNidever08
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


def roi_cut(xcut, ycut, x, y):
    """Use cuts in a 2D plane to select points from arrays.

    Parameters
    ----------
    xcut : numpy array
        Array of x-values for the cut.
    ycut : numpy array
        Array of y-values for the cut.
    x : numpy array or list
        Array of x-values that should be cut.
    y : numpy array or list
        Array of y-values that should be cut.

    Returns
    -------
    ind : numpy array
       The indices of values OUTSIDE the cut.
    cutind :
       The indices of values INSIDE the cut.

    """

    from matplotlib.path import Path

    tupVerts = list(zip(xcut, ycut))

    points = numpy.vstack((x, y)).T

    p = Path(tupVerts)  # make a polygon
    inside = p.contains_points(points)

    (ind,) = numpy.where(~inside)
    (cutind,) = numpy.where(inside)

    return ind, cutind


class MWM_MagCloud_AGB_APOGEE(BaseCarton):
    """MWM Magellanic clouds AGBs.

    Definition:

    Select AGB targets in the Magellanic Clouds using Gaia DR3 photometry and astrometry.
    Selection based on color cuts, proper motion cuts and parallax cuts. See wiki
    for David Nidever's code.

    """

    name = "mwm_magcloud_agb_apogee"
    mapper = "MWM"
    category = "science"
    program = "mwm_magcloud"
    instrument = "APOGEE"
    priority = 2818
    can_offset = True

    def build_query(self, version_id, query_region=None):
        # Parallax cut
        parallax_cut = ~(
            (Gaia_DR3.parallax > 0) & ((Gaia_DR3.parallax + 0.025) / Gaia_DR3.parallax_error > 5)
        )

        # Rough cuts. Just to reduce the number of rows returned for post-process.
        bp_rp = Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag
        colour_cut = (Gaia_DR3.phot_g_mean_mag <= self.parameters["gaiag_rough"]) & (
            bp_rp >= self.parameters["bp_rp_rough"]
        )

        pm_cut = (
            (Gaia_DR3.pmra >= self.parameters["pmra_rough"][0])
            & (Gaia_DR3.pmra <= self.parameters["pmra_rough"][1])
            & (Gaia_DR3.pmdec >= self.parameters["pmdec_rough"][0])
            & (Gaia_DR3.pmdec <= self.parameters["pmdec_rough"][1])
        )

        ra, dec, radius = self.parameters["astro_rough"]
        astro_cut = fn.q3c_radial_query(Gaia_DR3.ra, Gaia_DR3.dec, ra, dec, radius)

        # Cadence is calculated as h2exp from mwm_tess_rgb with sn=45.
        SN = 45.0
        exptime = 15.0

        time = 60 * (fn.POW(SN, 2) / 10000) * fn.POW(10, (0.4 * (TwoMassPSC.h_m - 11)))
        nexp = fn.ROUND(time / exptime)
        # We use bright_1x1 for both nexp=0 and nexp=1.
        #
        # For nexp == 1, we use bright_1x1 since
        # there is no cadence bright_flexible_1x1.
        #
        # For syntax of peewee.Case(), search for 'CASE statement'
        # in the below link.
        # https://docs.peewee-orm.com/en/latest/peewee/api.html
        cadence = peewee.Case(
            None,
            (
                (nexp == 0, "bright_1x1"),
                (nexp == 1, "bright_1x1"),
            ),
            fn.CONCAT("bright_flexible_", nexp.cast("text"), "x1"),
        )

        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                Gaia_DR3.l,
                Gaia_DR3.b,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                Gaia_DR3.parallax,
                Gaia_DR3.parallax_error,
                TwoMassPSC.h_m,
                TwoMassPSC.j_m,
                TwoMassPSC.k_m,
                cadence.alias("cadence"),
            )
            .join(Gaia_DR3)
            .join_from(
                CatalogToGaia_DR3,
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC)
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < self.parameters["h_lim"],
                parallax_cut,
                colour_cut,
                pm_cut,
                astro_cut,
            )
        )

        return query

    def post_process(self, model, **kwargs):
        """Runs post-process."""

        data = pandas.read_sql(f"SELECT * from {self.path}", self.database)

        # Calculate Magellanic Stream coordinates
        coords = SkyCoord(
            ra=data.ra.values,
            dec=data.dec.values,
            unit="deg",
            frame="icrs",
            pm_ra_cosdec=data.pmra.values * mas / yr,
            pm_dec=data.pmdec.values * mas / yr,
        )
        mcoo = coords.transform_to(MagellanicStreamNidever08)

        data["pml_ms"] = mcoo.pm_L_cosB.value
        data["pmb_ms"] = mcoo.pm_B.value
        data["mlat"] = mcoo.B.value
        data["mlon"] = mcoo.L.value

        # Make sure none of the needed quantities are NaN
        good = (
            numpy.isfinite(data["pmra"])
            & numpy.isfinite(data["pmdec"])
            & numpy.isfinite(data["parallax"])
            & numpy.isfinite(data["j_m"])
            & numpy.isfinite(data["h_m"])
            & numpy.isfinite(data["k_m"])
        )
        data = data[good]

        # Proper motion cut.
        pmdist = numpy.sqrt((data["pml_ms"] - 1.8) ** 2 + (data["pmb_ms"] - 0.40) ** 2)
        gdpm = pmdist < self.parameters["pmdist"]
        data = data.loc[gdpm]

        mlon_a = self.parameters["mlon_a"]
        mlon_b = self.parameters["mlon_b"]

        # CMD cut
        # AGB-O and AGB-C
        # Apply separate cuts to LMC and SMC

        # -- LMC --
        jkcut_lmc = numpy.array(self.parameters["jkcut_lmc"])
        hcut_lmc = numpy.array(self.parameters["hcut_lmc"])
        glmc = data["mlat"] > (mlon_a * data["mlon"] + mlon_b)
        data_lmc = data.loc[glmc]
        _, cutind = roi_cut(
            jkcut_lmc, hcut_lmc, data_lmc["j_m"] - data_lmc["k_m"], data_lmc["h_m"]
        )
        data_lmc = data_lmc.iloc[cutind]

        # -- SMC --
        jkcut_smc = jkcut_lmc + self.parameters["jkcut_smc_off"]
        hcut_smc = hcut_lmc + self.parameters["hcut_smc_off"]
        gsmc = data["mlat"] <= (mlon_a * data["mlon"] + mlon_b)
        data_smc = data.loc[gsmc]
        _, cutind = roi_cut(
            jkcut_smc, hcut_smc, data_smc["j_m"] - data_smc["k_m"], data_smc["h_m"]
        )
        data_smc = data_smc.iloc[cutind]

        valid_cids = numpy.hstack([data_lmc.catalogid.values, data_smc.catalogid.values])

        (
            model.update({model.selected: False})
            .where(model.catalogid.not_in(valid_cids.tolist()))
            .execute()
        )

        return super().post_process(model, **kwargs)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date: 2017-12-05 12:01:21


class TargetSelectionError(Exception):
    """A custom core TargetSelection exception"""

    def __init__(self, message=None):
        message = "There has been an error." if not message else message
        super(TargetSelectionError, self).__init__(message)


class TargetSelectionNotImplemented(TargetSelectionError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):
        message = "This feature is not implemented yet." if not message else message
        super(TargetSelectionNotImplemented, self).__init__(message)


class TargetSelectionMissingDependency(TargetSelectionError):
    """A custom exception for missing dependencies."""

    pass


class XMatchError(TargetSelectionError):
    """An error during cross-matching."""


class TargetSelectionWarning(Warning):
    """Base warning for TargetSelection."""


class TargetSelectionUserWarning(UserWarning, TargetSelectionWarning):
    """The primary warning class."""

    pass


class TargetSelectionDeprecationWarning(TargetSelectionUserWarning):
    """A warning for deprecated features."""

    pass


class TargetSelectionImportWarning(TargetSelectionUserWarning):
    """Warning for import problems."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta
# @Date: 2023-02-28
# @Filename: mwm_galactic.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import random

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


# The below MWM_Galactic_Core_apogee_Carton carton is obsolete.
# Do not run this carton for v1.0.
# However, do not delete this carton since
# this carton is referred to by the MWM_Galactic_Core_Dist_apogee_Carton
class MWM_Galactic_Core_apogee_Carton(BaseCarton):
    """Galactic Genesis carton.

    Definition: Selection of all IR-bright, red stars - vast majority are red
    giants; follows the density distribution of the MW (concentrated in the
    plane and bulge). Select sources brighter than H<11 AND ((G-H) > 3.5 OR
    Gaia non-detection).

    v0.5
    Pseudo-query:

        SELECT catalog.catalogid, catalog.ra, catalog.ra, catalog.dec,
            catalog.pmra, catalog.pmdec, catalog.epoch, twomass_psc.h_m,
            gaia_dr2_source.phot_g_mean_mag
        FROM catalog
        JOIN catalog_to_TIC, TIC, gaia_dr2_source, gaia_clean, twomass_psc
        WHERE h_m < 11
            AND (ph_qual[1] = 'A' OR ph_qual[1] = 'B')
            AND gal_contam=0
            AND cc_flg[1]=0
            AND (rd_flg[1] > 0 AND rd_flg[1] <= 3)
            AND [(phot_g_mean_mag-h_m) > 3.5  OR NOT_EXISTS(phot_g_mean_mag)]

    v1.0
    Shorthand name: mwm_galactic_core_apogee
    Existing carton code:
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_galactic.py

    Simplified Description of selection criteria:
    Selection of all IR-bright, red stars – vast majority are red giants;
     follows the density distribution of the MW (concentrated in the plane and bulge).
     Select sources brighter than H<11 AND ((G-H) > 3.5 OR Gaia non-detection).
    Gaia DR2 parameters to be converted to Gaia DR3: position information, phot_g_mean_mag,

    Pseudo-code: Same as v0.5

    SELECT catalog.catalogid, gaia_dr3.source_id, gaia_dr3.ra,  gaia_dr3.dec,
        gaia_dr3.pmra, gaia_dr3.pmdec,  twomass_psc.h_m, twomass_psc.designation,
        gaia_dr3.phot_g_mean_mag
    FROM catalog JOIN catalog_to_twomass_psc, catalog_to_gaia_dr3,  gaia_dr3,  twomass_psc
    WHERE
    h_m < 11 AND
    (ph_qual[1] = 'A' OR ph_qual[1] = 'B') AND
    gal_contam=0 AND
    cc_flg[1]=0 AND
    (rd_flg[1] > 0 AND rd_flg[1] <= 3)
    AND [(phot_g_mean_mag-h_m) > 3.5 OR NOT_EXISTS(phot_g_mean_mag)]

    Joins: TwoMassPSC, catalog_to_twomass_psc, catalog_to_gaia_dr3, gaia_dr3

    Return columns: Gaia E/DR3 source_id, ra, dec, parallax,
     Gaia E/DR3 G mag, Gaia E/DR3 BP mag, Gaia E/DR3 RP mag, J mag,
      H mag K mag, catalogid, 2MASS designation, 2MASS ph_qual, cc_flg, rd_flg, gal_contam

    Metadata:
    unchanged
    can_offset=TRUE
    Lead contact: Jonathan Bird

    """

    name = "mwm_galactic_core_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2710
    program = "mwm_galactic"
    mapper = "MWM"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToTwoMassPSC.select(
                CatalogToTwoMassPSC.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                Gaia_DR3.phot_g_mean_mag,
                TwoMassPSC.h_m,
                TwoMassPSC.designation,
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(CatalogToTwoMassPSC)
            .join(
                CatalogToGaia_DR3,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToTwoMassPSC.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(
                Gaia_DR3,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id),
            )
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                TwoMassPSC.h_m < 11,
                (peewee.fn.substr(TwoMassPSC.ph_qual, 2, 1) == "A")
                | (peewee.fn.substr(TwoMassPSC.ph_qual, 2, 1) == "B"),
                TwoMassPSC.gal_contam == 0,
                peewee.fn.substr(TwoMassPSC.cc_flg, 2, 1) == "0",
                (peewee.fn.substr(TwoMassPSC.rd_flg, 2, 1).cast("integer") > 0)
                & (peewee.fn.substr(TwoMassPSC.rd_flg, 2, 1).cast("integer") <= 3),
                ((Gaia_DR3.phot_g_mean_mag - TwoMassPSC.h_m) > 3.5)
                | (Gaia_DR3.phot_g_mean_mag >> None),
            )
        )
        # above condition (Gaia_DR3.phot_g_mean_mag >> None) ensures that
        # that we get the rows from the left outer join

        if query_region:
            query = query.join_from(CatalogToTwoMassPSC, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_Galactic_Core_Dist_apogee_Carton(BaseCarton):
    """Galactic Genesis carton.

    Definition: Selection of all IR-bright, red stars - vast majority are red
    giants; follows the density distribution of the MW (concentrated in the
    plane and bulge). Select sources brighter than H<11 AND ((G-H) > 5 OR
    Gaia non-detection).

    v0.5 only had mwm_galactic_core
    (there was no mwm_galactic_core_dist_apogee in v0.5)

    v1.0
    Shorthand name: mwm_galactic_core_dist_apogee
    This is a new carton. Code should be put in mwm_galactic.py.
    https://github.com/sdss/target_selection/blob/main/python/target_selection/cartons/mwm_galactic.py

    Simplified Description of selection criteria:
    Selection of all IR-bright, red stars – vast majority are red giants;
     follows the density distribution of the MW (concentrated in the plane and bulge).
     Select sources brighter than H<11 AND ((G-H) > 5 OR Gaia non-detection).
    Gaia DR2 parameters to be converted to Gaia DR3: position information, phot_g_mean_mag,

    Pseudo-code: Same as v1.0 mwm_galactic_core.

    SELECT catalog.catalogid, gaia_dr3.source_id, gaia_dr3.ra,  gaia_dr3.dec,
        gaia_dr3.pmra, gaia_dr3.pmdec,  twomass_psc.h_m, twomass_psc.designation,
        gaia_dr3.phot_g_mean_mag
    FROM catalog JOIN catalog_to_twomass_psc, catalog_to_gaia_dr3,  gaia_dr3,  twomass_psc
    WHERE
    h_m < 11 AND
    (ph_qual[1] = 'A' OR ph_qual[1] = 'B') AND
    gal_contam=0 AND
    cc_flg[1]=0 AND
    (rd_flg[1] > 0 AND rd_flg[1] <= 3)

    Above  is same as for v1.0 mwm_galactic_core_apogee.
    Below is additional for v1.0 mwm_galactic_core_dist_apogee.
    The below condition on g.phot_g_mean_mag is different from the condition
    in the mwm_galactic_core carton.

    tmass.h_m < 11 AND
    (

    ( gaia.parallax < 100 * power(10, 0.2 * (-2.5 - tmass.h_m)) ) OR
    (g.phot_g_mean_mag - tmass.h_m > 5) OR
    (NO gaia.phot_g_mean_mag)
    )

    Joins: TwoMassPSC, catalog_to_twomass_psc, catalog_to_gaia_dr3, gaia_dr3

    Return columns: Gaia E/DR3 source_id, ra, dec, parallax,
     Gaia E/DR3 G mag, Gaia E/DR3 BP mag, Gaia E/DR3 RP mag, J mag,
      H mag K mag, catalogid, 2MASS designation, 2MASS ph_qual, cc_flg, rd_flg, gal_contam

    Metadata:
    unchanged
    can_offset=TRUE
    Lead contact: Jonathan Bird

    """

    name = "mwm_galactic_core_dist_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2710
    program = "mwm_galactic"
    mapper = "MWM"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToTwoMassPSC.select(
                CatalogToTwoMassPSC.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                Gaia_DR3.parallax,
                Gaia_DR3.phot_g_mean_mag,
                TwoMassPSC.h_m,
                TwoMassPSC.designation,
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .switch(CatalogToTwoMassPSC)
            .join(
                CatalogToGaia_DR3,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToTwoMassPSC.catalogid == CatalogToGaia_DR3.catalogid),
            )
            .join(
                Gaia_DR3,
                peewee.JOIN.LEFT_OUTER,
                on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id),
            )
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                TwoMassPSC.h_m < 11,
                (peewee.fn.substr(TwoMassPSC.ph_qual, 2, 1) == "A")
                | (peewee.fn.substr(TwoMassPSC.ph_qual, 2, 1) == "B"),
                TwoMassPSC.gal_contam == 0,
                peewee.fn.substr(TwoMassPSC.cc_flg, 2, 1) == "0",
                (peewee.fn.substr(TwoMassPSC.rd_flg, 2, 1).cast("integer") > 0)
                & (peewee.fn.substr(TwoMassPSC.rd_flg, 2, 1).cast("integer") <= 3),
                (Gaia_DR3.parallax < 100 * peewee.fn.power(10, 0.2 * (-2.5 - TwoMassPSC.h_m)))
                | ((Gaia_DR3.phot_g_mean_mag - TwoMassPSC.h_m) > 5)
                | (Gaia_DR3.phot_g_mean_mag >> None),
            )
        )

        # above condition (Gaia_DR3.phot_g_mean_mag >> None) ensures that
        # that we get the rows from the left outer join

        if query_region:
            query = query.join_from(CatalogToTwoMassPSC, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_Galactic_Core_Dist_apogee_sparse_Carton(MWM_Galactic_Core_Dist_apogee_Carton):
    """mwm_galactic_core_dist_apogee_sparse
    Short description: Sparse-sampled Galactic Genesis sample.
    First do an initial selection as in mwm_galactic_core_dist_apogee and order by catalogid.
    Then randomly select exactly 2/3 of the sources using a fixed random seed.
    Metadata:
    Priority: 2710
    Cadence: bright_1x1
    Instrument: APOGEE
    can_offset = True
    Lead contact: Michael Blanton
    """

    name = "mwm_galactic_core_dist_apogee_sparse"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2710
    program = "mwm_galactic"
    mapper = "MWM"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        return query

    def post_process(self, model):
        cursor = self.database.execute_sql(
            "update sandbox.temp_mwm_galactic_core_dist_apogee_sparse " + "set selected = false;"
        )

        cursor = self.database.execute_sql(
            "select count(1) from " + " sandbox.temp_mwm_galactic_core_dist_apogee_sparse ;"
        )

        output = cursor.fetchall()

        total_num_rows = output[0][0]

        # This selected_fraction must be the same as in
        # mwm_galactic_core_dist_apogee_extra
        selected_fraction = 2.0 / 3.0
        selected_num_rows = int(selected_fraction * total_num_rows)

        b = [True] * selected_num_rows
        c = [False] * (total_num_rows - selected_num_rows)
        is_selected = b + c

        # This random seed must be the same as in
        # mwm_galactic_core_dist_apogee_extra
        random.seed(6789)
        random.shuffle(is_selected)

        # The below "order by catalogid" ensures that the random selection
        # further below gives the same result every time we run this carton.
        cursor = self.database.execute_sql(
            "select catalogid from "
            + " sandbox.temp_mwm_galactic_core_dist_apogee_sparse "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]

            # This condition is opposite of the condition in
            # mwm_galactic_core_dist_apogee_extra
            if is_selected[i] is True:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_galactic_core_dist_apogee_sparse "
                    + " set selected = true "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )


class MWM_Galactic_Core_Dist_apogee_extra_Carton(MWM_Galactic_Core_Dist_apogee_Carton):
    """mwm_galactic_core_dist_apogee_extra
    Short description: Complement of above carton mwm_galactic_core_dist_apogee_sparse.
    First do an initial selection as in mwm_galactic_core_dist_apogee and order by catalogid.
    Then select those targets which were NOT selected in
    above carton mwm_galactic_core_dist_apogee_sparse.
    Metadata:
    Priority: 2711
    Cadence: bright_1x1
    Instrument: APOGEE
    can_offset = True
    Lead contact: Michael Blanton
    """

    name = "mwm_galactic_core_dist_apogee_extra"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    priority = 2711
    program = "mwm_galactic"
    mapper = "MWM"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = super().build_query(version_id, query_region)
        return query

    def post_process(self, model):
        cursor = self.database.execute_sql(
            "update sandbox.temp_mwm_galactic_core_dist_apogee_extra " + "set selected = false;"
        )

        cursor = self.database.execute_sql(
            "select count(1) from " + " sandbox.temp_mwm_galactic_core_dist_apogee_extra ;"
        )

        output = cursor.fetchall()

        total_num_rows = output[0][0]

        # This selected_fraction must be the same as in
        # mwm_galactic_core_dist_apogee_sparse
        selected_fraction = 2.0 / 3.0
        selected_num_rows = int(selected_fraction * total_num_rows)

        b = [True] * selected_num_rows
        c = [False] * (total_num_rows - selected_num_rows)
        is_selected = b + c

        # This random seed must be the same as in
        # mwm_galactic_core_dist_apogee_sparse
        random.seed(6789)
        random.shuffle(is_selected)

        # The below "order by catalogid" ensures that the random selection
        # further below gives the same result every time we run this carton.
        cursor = self.database.execute_sql(
            "select catalogid from "
            + " sandbox.temp_mwm_galactic_core_dist_apogee_extra "
            + " order by catalogid;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]

            # select those sources which were not selected in
            # mwm_galactic_core_dist_apogee_sparse.
            # Hence, this condition is opposite of the condition in
            # mwm_galactic_core_dist_apogee_sparse
            if is_selected[i] is False:
                self.database.execute_sql(
                    " update sandbox.temp_mwm_galactic_core_dist_apogee_extra "
                    + " set selected = true "
                    + " where catalogid = "
                    + str(current_catalogid)
                    + ";"
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Zach Way (zway1@gsu.edu)
# @Date: 2024-04-23
# @Filename: append_to_sdss_id.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os
from datetime import date

import peewee
from peewee import JOIN, fn

from sdssdb.peewee.sdss5db.catalogdb import Catalog, database
from sdssdb.peewee.sdss5db.targetdb import Target

from .create_catalogidx_to_catalogidy import (
    MetaXMatch,
    TempMatch,
    UniqueMatch,
    create_unique_from_region,
)


class TempCatalogidV21(peewee.Model):
    """Model for a temporary table of all catalogids within a version."""

    pk = peewee.PrimaryKeyField()
    catalogid21 = peewee.BigIntegerField(index=True, null=False)

    class Meta:
        database = database
        schema = "sandbox"
        table_name = "temp_catalogid_v21"


class TempCatalogidV25(peewee.Model):
    """Model for a temporary table of all catalogids within a version."""

    pk = peewee.PrimaryKeyField()
    catalogid25 = peewee.BigIntegerField(index=True, null=False)

    class Meta:
        database = database
        schema = "sandbox"
        table_name = "temp_catalogid_v25"


class TempCatalogidV31(peewee.Model):
    """Model for a temporary table of all catalogids within a version."""

    pk = peewee.PrimaryKeyField()
    catalogid31 = peewee.BigIntegerField(index=True, null=False)

    class Meta:
        database = database
        schema = "sandbox"
        table_name = "temp_catalogid_v31"


class SdssIdStacked(peewee.Model):
    """Model for catalogdb sdss_id_stacked"""

    sdss_id = peewee.BigAutoField(primary_key=True)
    catalogid21 = peewee.BigIntegerField()
    catalogid25 = peewee.BigIntegerField()
    catalogid31 = peewee.BigIntegerField()
    ra_sdss_id = peewee.DoubleField()
    dec_sdss_id = peewee.DoubleField()
    last_updated = peewee.DateField()

    class Meta:
        database = database
        schema = "sandbox"  # "catalogdb"
        table_name = "sdss_id_stacked"


class SdssIdFlat(peewee.Model):
    """Model for catalogdb sdss_id_flat"""

    sdss_id = peewee.BigIntegerField()
    catalogid = peewee.BigIntegerField()
    version_id = peewee.SmallIntegerField()
    ra_sdss_id = peewee.DoubleField()
    dec_sdss_id = peewee.DoubleField()
    n_associated = peewee.SmallIntegerField()
    ra_catalogid = peewee.DoubleField()
    dec_catalogid = peewee.DoubleField()
    rank = peewee.SmallIntegerField(null=False)
    pk = peewee.BigAutoField(primary_key=True)

    class Meta:
        database = database
        schema = "sandbox"  # "catalogdb"
        table_name = "sdss_id_flat"


class SdssIdStackedAddendum(peewee.Model):
    """Model for addendum to sdss_id_stacked"""

    catalogid21 = peewee.BigIntegerField(index=True)
    catalogid25 = peewee.BigIntegerField(index=True)
    catalogid31 = peewee.BigIntegerField(index=True)
    ra_sdss_id = peewee.DoubleField()
    dec_sdss_id = peewee.DoubleField()
    last_updated = peewee.DateField()

    class Meta:
        database = database
        schema = "sandbox"


class SdssIdFlatAddendum(peewee.Model):
    """Model for addendum to sdss_id_flat"""

    sdss_id = peewee.BigIntegerField()
    catalogid = peewee.BigIntegerField()
    version_id = peewee.SmallIntegerField()
    ra_sdss_id = peewee.DoubleField()
    dec_sdss_id = peewee.DoubleField()
    n_associated = peewee.SmallIntegerField(null=True)
    ra_catalogid = peewee.DoubleField(null=True)
    dec_catalogid = peewee.DoubleField(null=True)
    rank = peewee.SmallIntegerField(null=True)

    class Meta:
        database = database
        schema = "sandbox"
        table_name = "sdss_id_flat_addendum"


class AppendToTables:
    """Adds new objects to the sdss_id tables.

    This class automatically uses the correlation method MetaXMatch and
    then has methods to put these objects through the sdss_id process. This
    can be done for 1) a list of catalogids, 2) a catalog_to_? table in
    catalogdb, or 3) all of targetdb.target

    Examples of how to run this class for each case are included at the end of
    the file append_to_sdss_id.py

    Parameters
    ----------
    database : peewee:PostgresqlDatabase
        A `peewee:PostgresqlDatabase` to the database the tables to
        cross-match.
    individual_table : str
        The table to add to the crossmatch in the form of catalogdb.catalog_to_?
        For example individual_table="catalogdb.catalog_to_too_target"
    catalogid_list : list
        A list of catalogids from the database
    """

    def __init__(self, database, individual_table=None, catalogid_list=None):
        self.database = database
        self.individual_table = individual_table
        self.ind_table_clean = self.individual_table
        self.catalogid_list = catalogid_list
        self.dir_path = os.path.dirname(__file__)

        if self.catalogid_list is not None:
            config = {
                "version_ids_to_match": [21, 25, 31],
                "individual_xmatch_config": self.dir_path + "/config/individual_crossmatches.yml",
                "log_file": "catalogidx_to_catalogidy_from_list.log",
                "show_first": 20,
                "split_insert_nunmber": 100000,
                "database_options": {"enable_hashjoin": "false"},
                # "split_query": [['panstarrs1',522000000000000,5000000000000]],
                "catalogid_list": catalogid_list,
            }

        elif self.individual_table is not None:
            if "." in self.individual_table:
                self.ind_table_clean = self.individual_table.split(".")[-1]

            config = {
                "version_ids_to_match": [21, 25, 31],
                "individual_xmatch_config": self.dir_path + "/config/individual_crossmatches.yml",
                "log_file": f"catalogidx_to_catalogidy_{self.ind_table_clean}.log",
                "show_first": 20,
                "split_insert_nunmber": 100000,
                "database_options": {"enable_hashjoin": "false"},
                # "split_query": [['panstarrs1', 522000000000000, 5000000000000]],
                "individual_table": individual_table,
            }
        else:
            config = {
                "version_ids_to_match": [21, 25, 31],
                "individual_xmatch_config": self.dir_path + "/config/individual_crossmatches.yml",
                "log_file": "catalogidx_to_catalogidy_all_target.log",
                "show_first": 20,
                "split_insert_nunmber": 100000,
                "database_options": {"enable_hashjoin": "false"},
            }
        self.config = config

    def run_MetaXMatch(self, database):
        """This takes in an individual catalog_to_? table (labeled 'individual_table') or a
        list of catalogids and creates the catalogidx_to_catalogidy_?
        and catalogidx_to_catalogidy_?_unique tables.
        """
        metax = MetaXMatch(
            config_filename=None,
            database=database,
            from_yaml=False,
            from_dict=True,
            config_dict=self.config,
            outer_join_sdss_id=True,
        )

        metax.run()
        create_unique_from_region(metax.output_name)
        return metax.output_name

    def create_temp_catalogid_lists(self, database, output_name):
        """This method takes the newly created catalogidx_to_catalogidy table and
        creates three aggregate tables in the sandbox containing all of the possible
        catalogids to be matched. This is also where the process restricts the matching of
        catalogids to 1) those which are labeled a 'best' match and 2) catalogids which have an
        inter-crossmatch match on that catalogid's lead catalog"""

        TempMatch._meta.table_name = output_name
        UniqueMatch._meta.table_name = output_name + "_unique"

        v21_cid_query_x = (
            TempMatch.select(TempMatch.catalogidx.alias("catalogid21"))
            .join(
                Catalog,
                on=(
                    (Catalog.catalogid == TempMatch.catalogidx) & (Catalog.lead == TempMatch.lead)
                ),
            )
            .switch(TempMatch)
            .join(
                SdssIdStacked,
                join_type=JOIN.LEFT_OUTER,
                on=(TempMatch.catalogidx == SdssIdStacked.catalogid21),
            )
            .where((TempMatch.version_idx == 21) & (SdssIdStacked.catalogid21.is_null()))
        )

        v25_cid_query_x = (
            TempMatch.select(TempMatch.catalogidx.alias("catalogid25"))
            .join(
                Catalog,
                on=(
                    (Catalog.catalogid == TempMatch.catalogidx) & (Catalog.lead == TempMatch.lead)
                ),
            )
            .switch(TempMatch)
            .join(
                SdssIdStacked,
                join_type=JOIN.LEFT_OUTER,
                on=(TempMatch.catalogidx == SdssIdStacked.catalogid25),
            )
            .where((TempMatch.version_idx == 25) & (SdssIdStacked.catalogid25.is_null()))
        )

        v31_cid_query_x = (
            TempMatch.select(TempMatch.catalogidx.alias("catalogid31"))
            .join(
                Catalog,
                on=(
                    (Catalog.catalogid == TempMatch.catalogidx) & (Catalog.lead == TempMatch.lead)
                ),
            )
            .switch(TempMatch)
            .join(
                SdssIdStacked,
                join_type=JOIN.LEFT_OUTER,
                on=(TempMatch.catalogidx == SdssIdStacked.catalogid31),
            )
            .where((TempMatch.version_idx == 31) & (SdssIdStacked.catalogid31.is_null()))
        )

        v21_cid_query_y = (
            TempMatch.select(TempMatch.catalogidy.alias("catalogid21"))
            .join(
                Catalog,
                on=(
                    (Catalog.catalogid == TempMatch.catalogidx) & (Catalog.lead == TempMatch.lead)
                ),
            )
            .switch(TempMatch)
            .join(
                SdssIdStacked,
                join_type=JOIN.LEFT_OUTER,
                on=(TempMatch.catalogidy == SdssIdStacked.catalogid21),
            )
            .where((TempMatch.version_idy == 21) & (SdssIdStacked.catalogid21.is_null()))
        )

        v25_cid_query_y = (
            TempMatch.select(TempMatch.catalogidy.alias("catalogid25"))
            .join(
                Catalog,
                on=(
                    (Catalog.catalogid == TempMatch.catalogidx) & (Catalog.lead == TempMatch.lead)
                ),
            )
            .switch(TempMatch)
            .join(
                SdssIdStacked,
                join_type=JOIN.LEFT_OUTER,
                on=(TempMatch.catalogidy == SdssIdStacked.catalogid25),
            )
            .where((TempMatch.version_idy == 25) & (SdssIdStacked.catalogid25.is_null()))
        )

        v31_cid_query_y = (
            TempMatch.select(TempMatch.catalogidy.alias("catalogid31"))
            .join(
                Catalog,
                on=(
                    (Catalog.catalogid == TempMatch.catalogidx) & (Catalog.lead == TempMatch.lead)
                ),
            )
            .switch(TempMatch)
            .join(
                SdssIdStacked,
                join_type=JOIN.LEFT_OUTER,
                on=(TempMatch.catalogidy == SdssIdStacked.catalogid31),
            )
            .where((TempMatch.version_idy == 31) & (SdssIdStacked.catalogid31.is_null()))
        )

        for table in [TempCatalogidV21, TempCatalogidV25, TempCatalogidV31]:
            if table.table_exists():
                database.drop_tables([table])

        # self.database.create_tables([TempCatalogidV21, TempCatalogidV25, TempCatalogidV31])
        TempCatalogidV21.create_table()
        TempCatalogidV25.create_table()
        TempCatalogidV31.create_table()

        temp_catalodids_indexes = """ CREATE INDEX temp_catalogid_v21_catalogid21_idx ON
                                        sandbox.temp_catalogid_v21 (catalogid21);
                                      CREATE INDEX temp_catalogid_v25_catalogid25_idx ON
                                        sandbox.temp_catalogid_v25 (catalogid25);
                                      CREATE INDEX temp_catalogid_v31_catalogid31_idx ON
                                        sandbox.temp_catalogid_v31 (catalogid31); """
        self.database.execute_sql(temp_catalodids_indexes)

        TempCatalogidV21.insert_from(v21_cid_query_x, [TempCatalogidV21.catalogid21]).execute()
        TempCatalogidV21.insert_from(v21_cid_query_y, [TempCatalogidV21.catalogid21]).execute()
        TempCatalogidV25.insert_from(v25_cid_query_x, [TempCatalogidV25.catalogid25]).execute()
        TempCatalogidV25.insert_from(v25_cid_query_y, [TempCatalogidV25.catalogid25]).execute()
        TempCatalogidV31.insert_from(v31_cid_query_x, [TempCatalogidV31.catalogid31]).execute()
        TempCatalogidV31.insert_from(v31_cid_query_y, [TempCatalogidV31.catalogid31]).execute()

        if self.catalogid_list is not None:
            catid_input_query = (
                Catalog.select(Catalog.catalogid, Catalog.version_id)
                .join(
                    SdssIdFlat,
                    join_type=JOIN.LEFT_OUTER,
                    on=(SdssIdFlat.catalogid == Catalog.catalogid),
                )
                .where(Catalog.catalogid << self.catalogid_list)
                .where(SdssIdFlat.catalogid.is_null())
            )

            for row in catid_input_query:
                if row.version_id == 21:
                    TempCatalogidV21.insert(catalogid21=row.catalogid).execute()
                elif row.version_id == 25:
                    TempCatalogidV25.insert(catalogid25=row.catalogid).execute()
                elif row.version_id == 31:
                    TempCatalogidV31.insert(catalogid31=row.catalogid).execute()
        elif self.individual_table is not None:
            try:
                ind_table = self.database.models[self.individual_table]
            except:
                raise ValueError(
                    "Could not find the table in database model: " + self.individual_table
                )

            ind_table_input_query = (
                Catalog.select(Catalog.catalogid, Catalog.version_id)
                .join(ind_table, on=(Catalog.catalogid == ind_table.catalogid))
                .join(
                    SdssIdFlat,
                    join_type=JOIN.LEFT_OUTER,
                    on=(SdssIdFlat.catalogid == Catalog.catalogid),
                )
                .where(SdssIdFlat.catalogid.is_null())
            )

            for row in ind_table_input_query:
                if row.version_id == 21:
                    TempCatalogidV21.insert(catalogid21=row.catalogid).execute()
                elif row.version_id == 25:
                    TempCatalogidV25.insert(catalogid25=row.catalogid).execute()
                elif row.version_id == 31:
                    TempCatalogidV31.insert(catalogid31=row.catalogid).execute()
        else:
            target_input_query = (
                Catalog.select(Catalog.catalogid, Catalog.version_id)
                .join(Target, on=(Catalog.catalogid == Target.catalogid))
                .join(
                    SdssIdFlat,
                    join_type=JOIN.LEFT_OUTER,
                    on=(SdssIdFlat.catalogid == Catalog.catalogid),
                )
                .where(SdssIdFlat.catalogid.is_null())
            )

            for row in target_input_query:
                if row.version_id == 21:
                    TempCatalogidV21.insert(catalogid21=row.catalogid).execute()
                elif row.version_id == 25:
                    TempCatalogidV25.insert(catalogid25=row.catalogid).execute()
                elif row.version_id == 31:
                    TempCatalogidV31.insert(catalogid31=row.catalogid).execute()

    def create_sdss_id_stacked_addendum(self, database, output_name):
        """This method matches the previous catalogids with the 'sdss_id' process
        and places them in a temporary table in the sandbox (to be added later). At
        this point, the ra/dec_sdss_id are assigned, which are just the coordinates of
        the most recent catalogid in the match."""

        large_cte_query = f"""DROP TABLE IF EXISTS sandbox.sdss_id_stacked_addendum;
        CREATE TABLE sandbox.sdss_id_stacked_addendum as (
            with cat_ids21 as (select distinct tc.catalogid21, cat.lead from
                sandbox.temp_catalogid_v21 tc join catalog cat on cat.catalogid=tc.catalogid21),
            cat_ids25 as (select distinct tc.catalogid25, cat.lead from
                sandbox.temp_catalogid_v25 tc join catalog cat on cat.catalogid=tc.catalogid25),
            cat_ids31 as (select distinct tc.catalogid31, cat.lead from
                sandbox.temp_catalogid_v31 tc join catalog cat on cat.catalogid=tc.catalogid31),
            sq21_25 as (select cc.catalogidx, cc.catalogidy from sandbox.{output_name} cc
                    join catalog cat on cat.catalogid=cc.catalogidx and cat.lead=cc.lead
                    where cc.version_idx=21 and cc.version_idy=25),
            sq25_31 as (select cc.catalogidx, cc.catalogidy from sandbox.{output_name} cc
                    join catalog cat on cat.catalogid=cc.catalogidx and cat.lead=cc.lead
                    where cc.version_idx=25 and cc.version_idy=31),
            left21_to_25 as (select cat_ids21.catalogid21, sq21_25.catalogidy as catalogid25 from
                cat_ids21 left join sq21_25 on cat_ids21.catalogid21=sq21_25.catalogidx),
            add_outer25 as (select left21_to_25.catalogid21,
                cat_ids25.catalogid25 as catalogid25 from
                left21_to_25 full outer join cat_ids25 on
                left21_to_25.catalogid25=cat_ids25.catalogid25),
            left_25_to_31 as (select add_outer25.catalogid21,
                add_outer25.catalogid25, sq25_31.catalogidy as catalogid31 from
                add_outer25 left join sq25_31 on add_outer25.catalogid25=sq25_31.catalogidx),
            add_outer31 as (select left_25_to_31.catalogid21, left_25_to_31.catalogid25,
                cat_ids31.catalogid31 from
                left_25_to_31 full outer join cat_ids31 on
                left_25_to_31.catalogid31=cat_ids31.catalogid31)
            select * from add_outer31);"""
        self.database.execute_sql(large_cte_query)

        add_ra_dec_columns = """ ALTER TABLE sandbox.sdss_id_stacked_addendum
                                    ADD COLUMN ra_sdss_id double precision;
                                 ALTER TABLE sandbox.sdss_id_stacked_addendum
                                    ADD COLUMN dec_sdss_id double precision; """
        self.database.execute_sql(add_ra_dec_columns)

        add_last_updated_column = f""" ALTER TABLE sandbox.sdss_id_stacked_addendum
                                           ADD COLUMN last_updated DATE;
                                       UPDATE sandbox.sdss_id_stacked_addendum SET
                                           last_updated = '{str(date.today())}'"""
        self.database.execute_sql(add_last_updated_column)

        create_indexes = """ CREATE INDEX sdss_id_stacked_addendum_catalogid21_idx ON
                                sandbox.sdss_id_stacked_addendum (catalogid21);
                             CREATE INDEX sdss_id_stacked_addendum_catalogid25_idx ON
                                sandbox.sdss_id_stacked_addendum (catalogid25);
                             CREATE INDEX sdss_id_stacked_addendum_catalogid31_idx ON
                                sandbox.sdss_id_stacked_addendum (catalogid31); """
        self.database.execute_sql(create_indexes)

        SdssIdStackedAddendum._meta.table_name = "sdss_id_stacked_addendum"

        ra_dec_update31 = (
            SdssIdStackedAddendum.update(ra_sdss_id=Catalog.ra, dec_sdss_id=Catalog.dec)
            .from_(Catalog)
            .where(SdssIdStackedAddendum.catalogid31 == Catalog.catalogid)
            .where(SdssIdStackedAddendum.catalogid31.is_null(False))
        )

        ra_dec_update25 = (
            SdssIdStackedAddendum.update(ra_sdss_id=Catalog.ra, dec_sdss_id=Catalog.dec)
            .from_(Catalog)
            .where(SdssIdStackedAddendum.catalogid25 == Catalog.catalogid)
            .where(SdssIdStackedAddendum.catalogid31.is_null())
            .where(SdssIdStackedAddendum.catalogid25.is_null(False))
        )

        ra_dec_update21 = (
            SdssIdStackedAddendum.update(ra_sdss_id=Catalog.ra, dec_sdss_id=Catalog.dec)
            .from_(Catalog)
            .where(SdssIdStackedAddendum.catalogid21 == Catalog.catalogid)
            .where(SdssIdStackedAddendum.catalogid31.is_null())
            .where(SdssIdStackedAddendum.catalogid25.is_null())
            .where(SdssIdStackedAddendum.catalogid21.is_null(False))
        )

        ra_dec_update31.execute()
        ra_dec_update25.execute()
        ra_dec_update21.execute()

    def add_to_sdss_id_stacked(self, database):
        """This method adds the previously made sdss_id_stached_addendum to sdss_id_stacked"""

        sid_stacked_f = [
            SdssIdStacked.catalogid21,
            SdssIdStacked.catalogid25,
            SdssIdStacked.catalogid31,
            SdssIdStacked.ra_sdss_id,
            SdssIdStacked.dec_sdss_id,
            SdssIdStacked.last_updated,
        ]

        to_add = SdssIdStackedAddendum.select(
            SdssIdStackedAddendum.catalogid21,
            SdssIdStackedAddendum.catalogid25,
            SdssIdStackedAddendum.catalogid31,
            SdssIdStackedAddendum.ra_sdss_id,
            SdssIdStackedAddendum.dec_sdss_id,
            SdssIdStackedAddendum.last_updated,
        ).tuples()
        with self.database.atomic():
            SdssIdStacked.insert_many(to_add, fields=sid_stacked_f).execute()

    def create_sdss_id_flat_addendum(self, database):
        """This method takes the new sdss_id rows in sdss_id_stacked and
        adds them to a temporary table in the sandbox in the sdss_id_flat format.
        Here, the ra/dec_catalogid are populated with the specific values associated
        with the catalogid."""

        if SdssIdFlatAddendum.table_exists():
            self.database.drop_tables([SdssIdFlatAddendum])
            # self.database.create_tables([SdssIdFlatAddendum])
            SdssIdFlatAddendum.create_table()
        else:
            # self.database.create_tables([SdssIdFlatAddendum])
            SdssIdFlatAddendum.create_table()

        sid_flat_fields = [
            SdssIdFlatAddendum.sdss_id,
            SdssIdFlatAddendum.catalogid,
            SdssIdFlatAddendum.version_id,
            SdssIdFlatAddendum.ra_sdss_id,
            SdssIdFlatAddendum.dec_sdss_id,
        ]

        max_sdss_id = SdssIdFlat.select(fn.MAX(SdssIdFlat.sdss_id)).scalar()

        sid_flat_add_v21 = (
            SdssIdStacked.select(
                SdssIdStacked.sdss_id,
                SdssIdStacked.catalogid21,
                peewee.Value(21),
                SdssIdStacked.ra_sdss_id,
                SdssIdStacked.dec_sdss_id,
            )
            .where(SdssIdStacked.catalogid21.is_null(False))
            .where(SdssIdStacked.sdss_id > max_sdss_id)
            .tuples()
        )

        with self.database.atomic():
            SdssIdFlatAddendum.insert_many(sid_flat_add_v21, fields=sid_flat_fields).execute()

        sid_flat_add_v25 = (
            SdssIdStacked.select(
                SdssIdStacked.sdss_id,
                SdssIdStacked.catalogid25,
                peewee.Value(25),
                SdssIdStacked.ra_sdss_id,
                SdssIdStacked.dec_sdss_id,
            )
            .where(SdssIdStacked.catalogid25.is_null(False))
            .where(SdssIdStacked.sdss_id > max_sdss_id)
            .tuples()
        )

        with self.database.atomic():
            SdssIdFlatAddendum.insert_many(sid_flat_add_v25, fields=sid_flat_fields).execute()

        sid_flat_add_v31 = (
            SdssIdStacked.select(
                SdssIdStacked.sdss_id,
                SdssIdStacked.catalogid31,
                peewee.Value(31),
                SdssIdStacked.ra_sdss_id,
                SdssIdStacked.dec_sdss_id,
            )
            .where(SdssIdStacked.catalogid31.is_null(False))
            .where(SdssIdStacked.sdss_id > max_sdss_id)
            .tuples()
        )
        with self.database.atomic():
            SdssIdFlatAddendum.insert_many(sid_flat_add_v31, fields=sid_flat_fields).execute()

        cte = (
            SdssIdFlatAddendum.select(SdssIdFlatAddendum.catalogid, fn.COUNT("*").alias("ct"))
            .group_by(SdssIdFlatAddendum.catalogid)
            .cte("catid_n_associated", columns=("catalogid", "ct"))
        )

        (
            SdssIdFlatAddendum.update(n_associated=cte.c.ct)
            .from_(cte)
            .where(SdssIdFlatAddendum.catalogid == cte.c.catalogid)
            .with_cte(cte)
            .execute()
        )

        (
            SdssIdFlatAddendum.update(ra_catalogid=Catalog.ra, dec_catalogid=Catalog.dec)
            .from_(Catalog)
            .where(SdssIdFlatAddendum.catalogid == Catalog.catalogid)
            .execute()
        )

        ranked_values = SdssIdFlatAddendum.select(
            SdssIdFlatAddendum.catalogid,
            SdssIdFlatAddendum.sdss_id,
            fn.RANK()
            .over(
                partition_by=[SdssIdFlatAddendum.catalogid], order_by=[SdssIdFlatAddendum.sdss_id]
            )
            .alias("rank"),
        )

        (
            SdssIdFlatAddendum.update(rank=ranked_values.c.rank)
            .from_(ranked_values)
            .where(
                (SdssIdFlatAddendum.catalogid == ranked_values.c.catalogid)
                & (SdssIdFlatAddendum.sdss_id == ranked_values.c.sdss_id)
            )
            .execute()
        )

    def add_to_sdss_id_flat(self, database):
        """This method adds sdss_id_flat_addendum to sdss_id_flat"""

        sid_flat_f = [
            SdssIdFlat.sdss_id,
            SdssIdFlat.catalogid,
            SdssIdFlat.version_id,
            SdssIdFlat.ra_sdss_id,
            SdssIdFlat.dec_sdss_id,
            SdssIdFlat.n_associated,
            SdssIdFlat.ra_catalogid,
            SdssIdFlat.dec_catalogid,
            SdssIdFlat.rank,
        ]

        to_add = SdssIdFlatAddendum.select(
            SdssIdFlatAddendum.sdss_id,
            SdssIdFlatAddendum.catalogid,
            SdssIdFlatAddendum.version_id,
            SdssIdFlatAddendum.ra_sdss_id,
            SdssIdFlatAddendum.dec_sdss_id,
            SdssIdFlatAddendum.n_associated,
            SdssIdFlatAddendum.ra_catalogid,
            SdssIdFlatAddendum.dec_catalogid,
            SdssIdFlatAddendum.rank,
        ).tuples()

        with self.database.atomic():
            SdssIdFlat.insert_many(to_add, fields=sid_flat_f).execute()


"""
EXAMPLE OF ADDING A SERIES OF CATALOGIDS
foo = AppendToTables(database, catalogid_list=[list, of, integer, catalogids])
output_name = foo.run_MetaXMatch(database)
foo.create_temp_catalogid_lists(database, output_name)
foo.create_sdss_id_stacked_addendum(database, output_name)
foo.add_to_sdss_id_stacked(database)
foo.create_sdss_id_flat_addendum(database)
foo.add_to_sdss_id_flat(database)
"""

"""
EXAMPLE OF ADDING MISSING CATALOGIDS FROM A TABLE
foo = AppendToTables(database, individual_table='catalogdb.catalog_to_too_target')
output_name = foo.run_MetaXMatch(database)
foo.create_temp_catalogid_lists(database, output_name)
foo.create_sdss_id_stacked_addendum(database, output_name)
foo.add_to_sdss_id_stacked(database)
foo.create_sdss_id_flat_addendum(database)
foo.add_to_sdss_id_flat(database)
"""

"""
EXAMPLE OF ADDING MISSING CATALOGIDS FROM TARGETDB
foo = AppendToTables(database)
output_name = foo.run_MetaXMatch(database)
foo.create_temp_catalogid_lists(database, output_name)
foo.create_sdss_id_stacked_addendum(database, output_name)
foo.add_to_sdss_id_stacked(database)
foo.create_sdss_id_flat_addendum(database)
foo.add_to_sdss_id_flat(database)
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2021-09-23
# @Filename: ops_bright_stars.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    CatalogToTycho2,
    Gaia_DR3,
    TwoMassPSC,
    Tycho2,
)

from target_selection.cartons import BaseCarton
from target_selection.exceptions import TargetSelectionError


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py


class OPS_Gaia_Brightneighbors_Carton(BaseCarton):
    """6.1.1. Bright Gaia (G < 12) Stars
    Owner: Kevin Covey

    Shorthand name:
    ops_gaia_brightneighbors

    Simplified Description of selection criteria:
    "Select all objects from gaia DR3 with G < 13"

    Wiki page: NA

    Additional source catalogs needed: None

    Additional cross-matching needed:  None

    Return columns: Gaia id, Gaia RA, Gaia Dec, Gaia RA proper motion,
     Gaia Dec proper motion, G, BP, RP

    cadence options for these targets (list all options,
    even though no single target will receive more than one):
    Null (since this is a veto catalog,
    we want cadence, value, priority and instrument to all be Null).

    Notes:  Gaia magnitudes should be transformed onto the SDSS system
    for the targetdb.magnitude table in the standard way
    (including recording the appropriate value in the opt_prov field).

    """

    name = "ops_gaia_brightneighbors"
    category = "veto_location_boss"
    instrument = None
    cadence = None
    program = "ops"
    mapper = None
    priority = None
    can_offset = False

    # target_selection propagates the following columns if they are generated
    # during the query (with exactly the below name)
    # g, r, i, z, h, j, k, bp, rp gaia_g, optical_prov
    #
    # If any g,r, i, z, optical_prov are missing, then the code will
    # try to find them in SDSS, PS1, and Gaia.
    #
    # Hence, we use the names gaia_g, bp, rp below.

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.pmra.alias("gaia_dr3_pmra"),
                Gaia_DR3.pmdec.alias("gaia_dr3_pmdec"),
                Gaia_DR3.phot_g_mean_mag.alias("gaia_g"),
                Gaia_DR3.phot_bp_mean_mag.alias("bp"),
                Gaia_DR3.phot_rp_mean_mag.alias("rp"),
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                Gaia_DR3.phot_g_mean_mag < 13,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class OPS_Tycho2_Brightneighbors_Carton(BaseCarton):
    """6.1.2. Bright Tycho2 (VT < 13) Stars
    Owner: Kevin Covey

    Shorthand name:
    ops_tycho2_brightneighbors

    Simplified Description of selection criteria:
    "Select all objects from Tycho2 with VT < 13"

    Wiki page: NA

    Additional source catalogs needed: None

    Additional cross-matching needed:  None

    Return columns: Tycho2 id, Tycho2 RA, Tycho2 Dec,
    Tycho2 RA proper motion, Tycho2 Dec proper motion, VT, BT

    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    Null (since this is a veto catalog, we want cadence, value,
    priority and instrument to all be Null).

    Notes:  Tycho2 magnitudes VT, BT will be transformed to pseudo-gaia_g magnitudes
    calculated for the targetdb.magnitudes table
    using the transforms from Evans et al. (2018):
            G = VT - 0.02051 - 0.2706 * (BT - VT) +
            0.03394 * (BT - VT)^2 - 0.05937 * (BT - VT)^3
    all other magnitudes can be stored as 'null',
    and a new optical_prov entry should be used to indicate
    the source of these magnitudes.
    For the rows with the above transform, optical_prov = 'gaia_psfmag_tycho2a'.

    If VT is not null and BT is null then the Tycho2 VT magnitude will be
    transformed to pseudo gaia_g magnitudes calculated for
    the targetdb.magnitudes table using the ad-hoc transform:
                  G = VT - 1
     For the rows with the ad-hoc transform, optical_prov = 'gaia_psfmag_tycho2b'.
    """

    name = "ops_tycho2_brightneighbors"
    category = "veto_location_boss"
    instrument = None
    cadence = None
    program = "ops"
    mapper = None
    priority = None
    can_offset = False

    # The column tycho2.designation is the primary key of the table catalogdb.tycho2.
    # Hence below we use
    # on=(CatalogToTycho2.target_id == Tycho2.designation)
    #
    # We will not fill the g, r, i, z columns.
    # However, we will create these columns in the query here.
    # This is because if these columns exist then add_optical_magnitudes()
    # will return right away.
    # So add_optical_magnitudes() will not create the g, r, i, z, and
    # optical_prov columns and
    # it will not run queries to get g, r, i, z.
    #
    # We will fill the gaia_g and optical_prov columns in post_process().
    #
    # These columns must be part of the query as shown below.
    # They cannot be added later in post_process().

    def build_query(self, version_id, query_region=None):
        # In the below code,
        # in cast('text'), 'text' refers to the PostgreSQL column type 'text'
        # and
        # in cast('real'), 'real' refers to the PostgreSQL column type 'real'
        optical_prov = peewee.Value("sdss_psfmag_tycho2").cast("text")
        g = peewee.Value(None).cast("real")
        r = peewee.Value(None).cast("real")
        i = peewee.Value(None).cast("real")
        z = peewee.Value(None).cast("real")
        gaia_g = peewee.Value(None).cast("real")

        query = (
            CatalogToTycho2.select(
                CatalogToTycho2.catalogid,
                Tycho2.tycid,
                Tycho2.designation,
                Tycho2.ramdeg.alias("tycho2_ra"),
                Tycho2.demdeg.alias("tycho2_dec"),
                Tycho2.pmra.alias("tycho2_pmra"),
                Tycho2.pmde.alias("tycho2_pmde"),
                Tycho2.vtmag,
                Tycho2.btmag,
                optical_prov.alias("optical_prov"),
                g.alias("g"),
                r.alias("r"),
                i.alias("i"),
                z.alias("z"),
                gaia_g.alias("gaia_g"),
            )
            .join(Tycho2, on=(CatalogToTycho2.target_id == Tycho2.designation))
            .where(
                CatalogToTycho2.version_id == version_id,
                CatalogToTycho2.best >> True,
                Tycho2.vtmag < 13,
            )
        )

        if query_region:
            query = query.join_from(CatalogToTycho2, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model):
        """
        Compute new column gaia_g from tycho2 vtmag and btmag.
        """

        cursor = self.database.execute_sql(
            "select catalogid, vtmag, btmag from " + " sandbox.temp_ops_tycho2_brightneighbors ;"
        )

        output = cursor.fetchall()

        for i in range(len(output)):
            current_catalogid = output[i][0]
            vtmag = output[i][1]
            btmag = output[i][2]

            if vtmag is not None:
                if btmag is not None:
                    current_gaia_g = (
                        vtmag
                        - 0.02051
                        - 0.2706 * (btmag - vtmag)
                        + 0.03394 * (btmag - vtmag) ** 2
                        - 0.05937 * (btmag - vtmag) ** 3
                    )
                    current_optical_prov = "gaia_psfmag_tycho2a"
                else:
                    # Since btmag is None, we cannot use the above equation.
                    # Below equation sets gaia_g to a bright value
                    current_gaia_g = vtmag - 1
                    current_optical_prov = "gaia_psfmag_tycho2b"
            else:
                raise TargetSelectionError(
                    "error: " + "ops_tycho2_brightneighbors post_process(): " + "vtmag is None"
                )

            self.database.execute_sql(
                " update sandbox.temp_ops_tycho2_brightneighbors "
                + " set gaia_g = "
                + str(current_gaia_g)
                + " where catalogid = "
                + str(current_catalogid)
                + ";"
            )

            self.database.execute_sql(
                " update sandbox.temp_ops_tycho2_brightneighbors "
                + " set optical_prov = '"
                + current_optical_prov
                + "'"
                + " where catalogid = "
                + str(current_catalogid)
                + ";"
            )


class OPS_2MASS_PSC_Brightneighbors_Carton(BaseCarton):
    """6.2.  Bright 2MASS (H < 7) Point Sources
    Owner: Kevin Covey

    Shorthand name:
    ops_2mass_psc_brightneighbors

    Simplified Description of selection criteria:
    "Select all objects from the 2MASS Point Source Catalog with H < 7"

    Wiki page: NA

    Additional source catalogs needed: None

    Additional cross-matching needed:  None

    Return columns:  2MASS ID, 2MASS RA, 2MASS Dec, J, H, K

    cadence options for these targets
    (list all options, even though no single target will receive more than one):
    Null (since this is a veto catalog, we want cadence, value,
    priority and instrument to all be Null).
    """

    name = "ops_2mass_psc_brightneighbors"
    category = "veto_location_apogee"
    instrument = None
    cadence = None
    program = "ops"
    mapper = None
    priority = None
    can_offset = False

    def build_query(self, version_id, query_region=None):
        # We do not select pmra and pmdec below because
        # twomass_psc table does not have pmra and pmdec.
        query = (
            CatalogToTwoMassPSC.select(
                CatalogToTwoMassPSC.catalogid,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                TwoMassPSC.ra.alias("twomass_psc_ra"),
                TwoMassPSC.decl.alias("twomass_psc_dec"),
                TwoMassPSC.j_m.alias("twomass_psc_j_m"),
                TwoMassPSC.h_m.alias("twomass_psc_h_m"),
                TwoMassPSC.k_m.alias("twomass_psc_k_m"),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .where(
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                TwoMassPSC.h_m < 7,
            )
        )

        if query_region:
            query = query.join_from(CatalogToTwoMassPSC, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2023-04-21
# @Filename: mwm_bin_gaia.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    Gaia_DR3,
    Gaia_dr3_nss_two_body_orbit,
)

from target_selection.cartons import BaseCarton


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py


class MWM_bin_gaia_astb_apogee_Carton(BaseCarton):
    """5.1.21. mwm_bin_gaia_astb_apogee
    Shorthand name:  mwm_bin_gaia_astb_apogee
    Existing carton code: adql_query_mwm_bin_gaia_astb_apogee.txt (this is
    ADQL code that runs on the Gaia archive) Simplified Description of
    selection criteria:  binaries from gaiadr3.nss_two_body_orbit satisfying
    (period < 1000) & (8 < phot_g_mean_mag < 13) & ((nss_solution_type =
    'Orbital') or (nss.nss_solution_type = 'AstroSpectroSB1')).
    This gives 62,738 sources.

    Return columns:
     query for full carton (returns 62,738)
     select gs.source_id, gs.ra, gs.dec, gs.phot_g_mean_mag
     from gaiadr3.nss_two_body_orbit as nss,
      gaiadr3.gaia_source as gs
      where (nss.nss_solution_type = 'Orbital' or
      nss.nss_solution_type = 'AstroSpectroSB1') and
      gs.source_id = nss.source_id and gs.phot_g_mean_mag between 8 and 13
      and period < 1000
    Metadata:
    Priority: 2550-2559
    Cadence: 2 visits separated by at least 100 days
     – new cadence    cadence assessed to be incompatible with existing set.
      Plan to test implementation with simplified cadence
      (bright_2x1_long_v2  - and allow partial completion / extra epoch eligibility)
    Instrument: APOGEE
    can_offset = True
    Lead contact: Kareem El-Badry
    """

    name = "mwm_bin_gaia_astb_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_2x1_long_v2"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2550
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_dr3_nss_two_body_orbit.period,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(
                Gaia_dr3_nss_two_body_orbit,
                on=(Gaia_DR3.source_id == Gaia_dr3_nss_two_body_orbit.source_id),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "Orbital")
                | (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "AstroSpectroSB1"),
                Gaia_DR3.phot_g_mean_mag.between(8, 13),
                Gaia_dr3_nss_two_body_orbit.period < 1000,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_bin_gaia_astb_boss_Carton(BaseCarton):
    """5.1.22. mwm_bin_gaia_astb_boss
    Shorthand name:  mwm_bin_gaia_astb_boss
    Existing carton code: adql_query_mwm_bin_gaia_astb_boss.txt
    Simplified Description of selection criteria:  binaries from
    gaiadr3.nss_two_body_orbit satisfying (period < 1000) & (13 <
    phot_g_mean_mag < 16) & ((nss_solution_type = 'Orbital') or
    (nss.nss_solution_type = 'AstroSpectroSB1')).
    This gives 87,470 sources.

    Return columns:
     query for full carton (returns 87470)
     select gs.source_id, gs.ra,
     gs.dec, gs.phot_g_mean_mag from gaiadr3.nss_two_body_orbit as nss,
     gaiadr3.gaia_source as gs where (nss.nss_solution_type = 'Orbital' or
     nss.nss_solution_type = 'AstroSpectroSB1')and gs.source_id =
     nss.source_id and gs.phot_g_mean_mag between 13 and 16 and period < 1000
    Metadata:
    Priority: 2560-2569
    Cadence:
    2 visits separated by at least 100 days – new cadence    cadence
    assessed to be incompatible with existing set. Plan to test
    implementation with simplified cadence (bright_2x1 / dark_2x1 - and
    allow partial completion / extra epoch eligibility)
    Instrument: BOSS
    can_offset = True
    Lead contact: Kareem El-Badry
    """

    name = "mwm_bin_gaia_astb_boss"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_2x1_long_v2"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2560
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_dr3_nss_two_body_orbit.period,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(
                Gaia_dr3_nss_two_body_orbit,
                on=(Gaia_DR3.source_id == Gaia_dr3_nss_two_body_orbit.source_id),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "Orbital")
                | (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "AstroSpectroSB1"),
                Gaia_DR3.phot_g_mean_mag.between(13, 16),
                Gaia_dr3_nss_two_body_orbit.period < 1000,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_bin_gaia_sb_apogee_Carton(BaseCarton):
    """5.1.23. mwm_bin_gaia_sb_apogee
    Shorthand name:  mwm_bin_gaia_sb_apogee Existing carton code:
    adql_query_mwm_bin_gaia_sb_apogee.txt Simplified Description of
    selection criteria:  binaries from gaiadr3.nss_two_body_orbit satisfying
    (period < 1000) & (8 < phot_g_mean_mag < 11.5) & ((nss_solution_type =
    'SB1') or (nss_solution_type = 'SB1C')).
    This gives 51,810 sources.

    Return columns:
     query for full carton (returns 51810)
     select gs.source_id, gs.ra,
     gs.dec, gs.phot_g_mean_mag from gaiadr3.nss_two_body_orbit as nss,
     gaiadr3.gaia_source as gswhere ((nss.nss_solution_type = 'SB1') or
     (nss.nss_solution_type = 'SB1C'))and gs.source_id = nss.source_idand
     gs.phot_g_mean_mag between 8 and 11.5 and period < 1000
    Metadata:
    Priority: 2570-2579
    Cadence: bright_1x1
    Instrument: APOGEE
    can_offset = True
    Lead contact: Kareem El-Badry
    """

    name = "mwm_bin_gaia_sb_apogee"
    category = "science"
    instrument = "APOGEE"
    cadence = "bright_1x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2570
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_dr3_nss_two_body_orbit.period,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(
                Gaia_dr3_nss_two_body_orbit,
                on=(Gaia_DR3.source_id == Gaia_dr3_nss_two_body_orbit.source_id),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "SB1")
                | (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "SB1C"),
                Gaia_DR3.phot_g_mean_mag.between(8, 11.5),
                Gaia_dr3_nss_two_body_orbit.period < 1000,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_bin_gaia_sb_boss_Carton(BaseCarton):
    """5.1.24. mwm_bin_gaia_sb_boss
    Shorthand name:  mwm_bin_gaia_sb_boss Existing carton code:
    adql_query_mwm_bin_gaia_sb_boss.txt Simplified Description of selection
    criteria:  binaries from gaiadr3.nss_two_body_orbit satisfying (period <
    1000) & (11.5 < phot_g_mean_mag < 14) & ((nss_solution_type = 'SB1') or
    (nss_solution_type = 'SB1C')).
    This gives 111,852 sources.

    Return columns:
     query for full carton (returns 111,852)
     select gs.source_id, gs.ra,
     gs.dec, gs.phot_g_mean_mag from gaiadr3.nss_two_body_orbit as nss,
     gaiadr3.gaia_source as gswhere ((nss.nss_solution_type = 'SB1') or
     (nss.nss_solution_type = 'SB1C'))and gs.source_id = nss.source_idand
     gs.phot_g_mean_mag between 11.5 and 14 and period < 1000
    Metadata:
    Priority: 2580-2589
    Cadence: bright_1x1
    Instrument: BOSS
    can_offset = True
    Lead contact: Kareem El-Badry
    """

    name = "mwm_bin_gaia_sb_boss"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_1x1"
    program = "mwm_bin"
    mapper = "MWM"
    priority = 2580
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.phot_g_mean_mag,
                Gaia_dr3_nss_two_body_orbit.period,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .join(
                Gaia_dr3_nss_two_body_orbit,
                on=(Gaia_DR3.source_id == Gaia_dr3_nss_two_body_orbit.source_id),
            )
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "SB1")
                | (Gaia_dr3_nss_two_body_orbit.nss_solution_type == "SB1C"),
                Gaia_DR3.phot_g_mean_mag.between(11.5, 14),
                Gaia_dr3_nss_two_body_orbit.period < 1000,
            )
        )

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query
